classdef rpsIO < handle
    
    properties
        
        par;                             % parameters and settings
        
        log;                             % variable logs
        
        fig;                             % figures
        
        ser;                             % service variables

        netOC;                           % operant conditioning network

        OCstimuli;                       % vector of input stimuli for OC
                       
        currentModulation               = 0;  % mapping from reward/punishment to network modulatory input

        rewarded                        = 0;
        punished                        = 0;
        
        labels = {'rock', 'paper', 'scissor', 'none'};
                
      end
    
    methods
        
        function app = rpsIO()
            disp('Rock, paper, scissor');
 
            % #############################################################
            %% Parameter settings 
            %##############################################################
            app.par.seed                        = 4114;                    % random seed for pseudo number generator

            app.par.simTime                     = 10800;                   % max duration of the experiment in seconds
            app.par.samplingTime                = 0.2;                     % sampling time in seconds
                        
            app.par.nrOCStimuli                 = 9;                       % number of stimuli
            app.par.nrOCActions                 = 3;                       % number of outputs
                        
            app.par.positiveReward               = 1;                       % intensity of modulation/per step for reward episode
            app.par.negativeReward              = -0.5 * app.par.positiveReward;
                      
            app.par.refreshGraphsEvery          = 1;
            
            app.netOC = rpsNet(app.par.seed, app.par.nrOCStimuli, app.par.nrOCActions, app.par.samplingTime, app.par.simTime, 2);
            
            
            % #############################################################
            % runtime and service variables
            
            app.OCstimuli = zeros(app.par.nrOCStimuli,1);

            app.log.idleTime = zeros(app.par.simTime / app.par.samplingTime, 1);

            app.ser.activeOutputDuration            = 0;
            app.ser.plottingcounter                 = 0;
            app.ser.currentOCStimulus               = 0;
            app.ser.timeFromLastAction              = 0;
            app.ser.actionCompleted                 = 0;

            app.createWindow();
            set(app.fig.panel,'KeyPressFcn',@app.keypressed);
            
            path = '../data/';
            if ~isdir(path)
                path = '../../data/';
            end
            
            icubphrases = load([path filesep 'icub-phrases.mat']);
            
            for i = 1:8
                 app.ser.voiceUS{i} = audioplayer(icubphrases.phrasesUS(i,:),44100);
            end
            clear icubphrases;
            
            icubphrasesRPS = load([path filesep 'RPSvoice.mat']);
            
            for i =1:8
                app.ser.voiceAction{i} = audioplayer(icubphrasesRPS.phrase{i}(:,:),44100);
            end
            clear icubphrasesRPS;            
 
        end
        
        function running(app,src,evn)
            if get(src, 'Value')
                app.ser.currentStep = 2;

               % measuring time for real-time computation (see app.realTimeMargin() )
               app.ser.recordedTime = clock;

                while get(src, 'Value') && app.ser.currentStep < app.par.simTime/app.par.samplingTime
                    
                    % slow decay or reward with 1h time constant
                    app.par.positiveReward = app.par.positiveReward .* exp(-app.par.samplingTime/3600);
                   
                    app.currentModulation = 0;
                    if app.ser.actionCompleted == 1
                        %disp(['time from last action = ' num2str(app.ser.timeFromLastAction)]);
                        app.ser.timeFromLastAction = 0;
                    end
                    app.ser.actionCompleted = 0;
                    
                    % implemented only in rpsRobot*
                    app.newTurn();
                    
                    % fetching INPUT HERE and setting reward ##############
                    app.readStimuli();
                    
                    % UPDATE NETWORK HERE #################################
                    app.netOC.update(app.OCstimuli, app.currentModulation);
                    
                    % READ OUTPUT #########################################
                    app.readOutput();
                    
                    if app.ser.actionCompleted == 1
                        app.OCstimuli = zeros(9,1);
                        app.ser.currentOCStimulus = 0;
                    end
                    
                    app.refreshGraphs();
                    
                    app.realTimeMargin();
                    
                    app.ser.recordedTime = clock();
                    
                    app.ser.currentStep = app.ser.currentStep + 1;
                    app.ser.timeFromLastAction = app.ser.timeFromLastAction + 1;
                end
            else
                %run button released
                app.stopRunning();
            end
        end
        
        function readStimuli(app)
            % in che child class update the variables
            % -- app.OCstimuli which is a vector
            % of dimensikon app.par.nrOStimuli,1
            % and
            % -- app.currentModulation
        end
        
        function newTurn(app)
        end
        
        function readOutput(app)
            app.ser.whoWins = find(app.netOC.actions == 1, 1);
            if ~isempty(app.ser.whoWins)
                app.ser.activeOutputDuration = app.ser.activeOutputDuration + 1;
            end
            if app.ser.activeOutputDuration > 1/app.par.samplingTime
                app.ser.activeOutputDuration = 0;
                app.ser.actionCompleted = 1;
                disp(['Input' num2str(app.ser.currentOCStimulus) ' Action ' num2str(app.ser.whoWins)]);
                if isa(app,'rpsIOAutoSim')
                    try
                        play(app.ser.voiceAction{app.ser.whoWins + 4});
                    end
                end
                app.triggerRobotActions();
            end
        end
        
        function  realTimeMargin(app)
            app.log.idleTime(app.ser.currentStep) = app.par.samplingTime - etime(clock,app.ser.recordedTime);
            
            if app.log.idleTime(app.ser.currentStep) < 0
                %disp(['ATTENTION! REAL TIME LOSS OF '  num2str(abs(app.log.idleTime(app.ser.currentStep))) ' s']);
            end
        end
        
        function b = isrewarded(app)
            b = 0;
        end
        
        function b = ispunished(app)
            b = 0;
        end
        
        function robotRewardOC(app)
        end
        
        function robotPunishmentOC(app)
        end
        
        function triggerRobotActions(app)
        end
        
        function stopRunning(app)
            app.OCstimuli = zeros(9,1);
            app.ser.actionCompleted = 1;
        end
        
        function keypressed(app, src, evt)
            %disp(evt.Key);
        end
        
        
        
        function createWindow(app)
            app.fig.panel = figure('Position',[0 0 1200 800]);
            set(app.fig.panel,'Units','normalized');
            set(app.fig.panel,'Name','Output');
            
            app.fig.OCpanel = uipanel(app.fig.panel,'Position',[0.01 0.27 0.99 0.75]);
            
            app.fig.genPanel = uipanel(app.fig.panel,'Position',[0 0 1 0.25]);
            
            app.fig.OCsp{1} = subplot('Position',[0.05 0.75 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{2} = subplot('Position',[0.30 0.75 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{3} = subplot('Position',[0.55 0.75 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{4} = subplot('Position',[0.80 0.75 0.2 0.20],'Parent',app.fig.OCpanel);
            
            app.fig.OCsp{5} = subplot('Position',[0.05 0.52 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{6} = subplot('Position',[0.30 0.52 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{7} = subplot('Position',[0.55 0.52 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{8} = subplot('Position',[0.80 0.52 0.2 0.20],'Parent',app.fig.OCpanel);
            
            app.fig.OCsp{9} = subplot('Position',[0.05 0.3 0.2 0.20],'Parent',app.fig.OCpanel);
            
            
            app.fig.OCsp{10} = subplot('Position',[0.3 0.3 0.2 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{11} = subplot('Position',[0.07 0.05 0.4 0.20],'Parent',app.fig.OCpanel);
            
            app.fig.OCsp{12} = subplot('Position',[0.58 0.3 0.4 0.20],'Parent',app.fig.OCpanel);
            app.fig.OCsp{13} = subplot('Position',[0.58 0.05 0.4 0.20],'Parent',app.fig.OCpanel);
            
            app.fig.RTmonitor = subplot('Position',[0.38 0.1 0.25 0.8],'Parent',app.fig.genPanel);

            app.fig.OCsp{14} = subplot('Position',[0.70 0.1 0.25 0.8],'Parent',app.fig.genPanel);
            
            app.ser.runbutton = uicontrol('Style', 'togglebutton',...
                'String','Start/Stop','Min',0,'Max',1,...
                'Parent',app.fig.panel,'Units','normalized','Position', [0 0 0.1 0.05],...
                'FontSize', 12,'Callback',@app.running);
            set(app.fig.OCpanel,'Visible','on');
            
        end
        
        function refreshGraphs(app)
            
            if mod(app.ser.currentStep,app.par.refreshGraphsEvery/(app.par.samplingTime)) == 0
                
                app.ser.plottingcounter = app.ser.plottingcounter + 1;
                
                switch mod(app.ser.plottingcounter,15) + 1
                    case 1,
                        plot(app.netOC.log.synapsesStoO{1}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{1});
                        set(app.fig.OCsp{1},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{1},'S1 to Ox');
                        
                    case 2,
                        plot(app.netOC.log.synapsesStoO{2}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{2});
                        set(app.fig.OCsp{2},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{2},'S2 to Ox');
                        
                    case 3,
                        plot(app.netOC.log.synapsesStoO{3}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{3});
                        set(app.fig.OCsp{3},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{3},'S3 to Ox');
                    case 4,
                        plot(app.netOC.log.synapsesStoO{4}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{4});
                        set(app.fig.OCsp{4},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{4},'S4 to Ox');
                    case 5,
                        plot(app.netOC.log.synapsesStoO{5}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{5});
                        set(app.fig.OCsp{5},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{5},'S5 to Ox');
                    case 6,
                        plot(app.netOC.log.synapsesStoO{6}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{6});
                        set(app.fig.OCsp{6},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{6},'S6 to Ox');
                    case 7,
                        plot(app.netOC.log.synapsesStoO{7}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{7});
                        set(app.fig.OCsp{7},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{7},'S7 to Ox');
                    case 8,
                        plot(app.netOC.log.synapsesStoO{8}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{8});
                        set(app.fig.OCsp{8},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{8},'S8 to Ox');
                    case 9,
                        plot(app.netOC.log.synapsesStoO{9}(1:(app.netOC.currentStep-1),:),'Parent',app.fig.OCsp{9});
                        set(app.fig.OCsp{9},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{9},'S9 to Ox');
                        
                    case 10,
                        plot([app.netOC.log.thresholdCor(max(1, app.netOC.currentStep - 499):app.netOC.currentStep)'...
                            app.netOC.log.thresholdDec(max(1, app.netOC.currentStep - 499):app.netOC.currentStep)'],'Parent',app.fig.OCsp{10});
                        ylabel(app.fig.OCsp{10},'Thresholds');
                    case 11,
                        plot(app.netOC.log.activityOG(max(1, app.netOC.currentStep - 499):app.netOC.currentStep,:),'Parent',app.fig.OCsp{11});
                        set(app.fig.OCsp{11},'Ylim',[-0.1 1]);
                        ylabel(app.fig.OCsp{11},'Output Groups');
                    case 12,
                        plot(app.netOC.log.modulation,'Parent',app.fig.OCsp{12});
                        set(app.fig.OCsp{12},'XLim',[1 app.netOC.currentStep-1],'YGrid','on');
                        ylabel(app.fig.OCsp{12},'Modulation');
                    case 13,
                        plot(sum(app.netOC.log.actions(max(1, app.netOC.currentStep - 299):app.netOC.currentStep,:),2),'Parent',app.fig.OCsp{13},'Line','none','Marker','.');
                        set(app.fig.OCsp{13},'Ylim',[0.5 app.netOC.par.nrOutputs + 0.5]);
                        set(app.fig.OCsp{13}','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
                        ylabel(app.fig.OCsp{13},'Action');
                    case 14,
                        plot(sum(app.netOC.log.stimuliRaw(max(1, app.netOC.currentStep - 299):app.netOC.currentStep,:),2),'Parent',app.fig.OCsp{14},'Marker','d','Line','--');
                        set(app.fig.OCsp{14},'Ylim',[0.5 app.netOC.par.nrStimuli + 0.5]);
                        set(app.fig.OCsp{14}','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
                    case 15,
                        plot(app.log.idleTime(max(1, app.ser.currentStep - 999):app.ser.currentStep),'Parent',app.fig.RTmonitor);
                        set(app.fig.RTmonitor,'YGrid','on','YLim',[-0.1 app.par.samplingTime]);
                        ylabel(app.fig.RTmonitor,'Real time left');
                end
                drawnow;
            end
        end
    end
end
