/**
 * @file ESpeakServer.cpp
 * @brief implementation of the ESpeakServerModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */ 

#include "iCub/ESpeakServerModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool ESpeakServerModule::configure(yarp::os::ResourceFinder &rf) {    
    /*
     * PLEASE remove useless comments when writing actual code. If needed then use Doxygen comments and tags.
     */

    /* Process all parameters from both command-line and .ini file */
    /* get the module name which will form the stem of all module port names */
    moduleName            = rf.check("name", 
                           Value("ESpeakServer"), 
                           "module name (string)").asString();

    /*
    * before continuing, set the module name before getting any other parameters, 
    * specifically the port names which are dependent on the module name
    */
    setName(moduleName.c_str());

    statusPortName        = "/";
    statusPortName       += getName(
                           rf.check("statusPort", 
                           Value("/status:o"),
                           "status port (string)").asString()
                           );

    /* get the threshold value */
    voice        = rf.check("voice",
                           Value("-v en+f1 -p 100"),
                           "voice param (string)").asString();

	text = "";
	speaking = false;
	receivednewtext = false;

    /* do all initialization here */
    /*
    * attach a port of the same name as the module (prefixed with a /) to the module
    * so that messages received from the port are redirected to the respond method
    */
    handlerPortName =  "/";
    handlerPortName += getName();         // use getName() rather than a literal
	handlerPortName += "/command"; 
    if (!handlerPort.open(handlerPortName.c_str())) {           
        cout << getName() << ": Unable to open port " << handlerPortName << endl;  
        return false;
    }
    attach(handlerPort); // attach to port
    attachTerminal(); // attach to terminal

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;  
		return false;
	}

    /* create the thread and pass pointers to the module parameters */
    dThread = new ESpeakServerThread(this);

    /* now start the thread to do the work */    
    dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

    return true ;     // let the RFModule know everything went well
                      // so that it will then run the module
}

bool ESpeakServerModule::interruptModule() {
    handlerPort.interrupt();
    return true;
}

bool ESpeakServerModule::close() {
    //close ports
    handlerPort.close();
    statusOut.interrupt();
    statusOut.close();

    /* stop the thread */
    dThread->stop();
    return true;
}

bool ESpeakServerModule::respond(const Bottle& command, Bottle& reply) {
  string helpMessage =  string(getName().c_str()) + 
                        " commands are: \n" +  
                        "help \n" + 
                        "quit \n" + 
                        "set thr <n> ... set the threshold \n" + 
                        "(where <n> is an integer number) \n";

  reply.clear(); 

  if (command.get(0).asString()=="quit") {
       reply.addString("quitting");
       return false;     
   }
   else if (command.get(0).asString()=="help") {
      cout << helpMessage;
      reply.addString("ok");
   } else if (command.get(0).asString()=="say") {
         std::string say = std::string(command.get(1).asString().c_str()); // set parameter value
		 cout << say << endl;
		 while (this->speaking)		sleep(0.01);
		 this->text = say;
		 this->receivednewtext = true;
         reply.addString("ok");
   } else if (command.get(0).asString()=="set") {
     	//TODO set voice etc 
   }
   return true;
}

bool ESpeakServerModule::openPorts() {
    if (!statusOut.open(statusPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }
	return true;
}

/* Called periodically every getPeriod() seconds */
bool ESpeakServerModule::updateModule() {

    return true;
}

double ESpeakServerModule::getPeriod() {
    /* module periodicity (seconds), called implicitly by myModule */    
    return 0.1;
}




