/**
 * @file MVITE.cpp
 * @brief implementation of the MVITEModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */

#include "iCub/MVITEModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool MVITEModule::configure(yarp::os::ResourceFinder &rf) {
	/*
	 * PLEASE remove useless comments when writing actual code. If needed then use Doxygen comments and tags.
	 */

	/* Process all parameters from both command-line and .ini file */
	/* get the module name which will form the stem of all module port names */
	moduleName =
			rf.check("name", Value("MVITE"), "module name (string)").asString();

	/*
	 * before continuing, set the module name before getting any other parameters,
	 * specifically the port names which are dependent on the module name
	 */
	setName(moduleName.c_str());

	/* now, get the rest of the parameters */
	/*
	 * get the robot name which will form the stem of the robot ports names
	 * and append the specific part and device required
	 */
	robotName =
			rf.check("robot", Value("icub"), "Robot name (string)").asString();

	bodyPart = rf.check("bodyPart", Value("right_arm"), "body part").asString();

	robotPortName = "/" + robotName + "/" + bodyPart;

	/* get the name of the input and output ports, automatically prefixing the module name by using getName() */
	outputPortName = "/";
	outputPortName += getName(
			rf.check("statusOutputPort", Value("/status:o"),
					"Output port (string)").asString());

	/* get the threshold value */
	alpha =
			rf.check("alpha", Value(1.0 / 10.0), "Key value (double)").asDouble();
	beta = rf.check("beta", Value(1.0 / 30.0), "Key value (double)").asDouble();
	noise = rf.check("noise", Value(0.1), "Key value (double)").asDouble();
	pause = true;

	/* do all initialization here */
	/*
	 * attach a port of the same name as the module (prefixed with a /) to the module
	 * so that messages received from the port are redirected to the respond method
	 */
	handlerPortName = "/";
	handlerPortName += getName(); // use getName() rather than a literal
	handlerPortName += "/command";
	if (!handlerPort.open(handlerPortName.c_str())) {
		cout << getName() << ": Unable to open port " << handlerPortName
				<< endl;
		return false;
	}
	attach(handlerPort); // attach to port
	attachTerminal(); // attach to terminal

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;
		return false;
	}

	//connect to robot
	success = openRobotInterface();
	if (!success) {
		cout << getName() << ": Unable to connect to robot " << robotName
				<< endl;
		return false;
	}

	//init vars
	success = this->init();
	if (!success) {
		cout << getName() << ": Unable to initialize (reading encoders)"
				<< endl;
		return false;
	}

	/* create the thread and pass pointers to the module parameters */
	dThread = new MVITEThread(this);

	/* now start the thread to do the work */
	dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

	return true; // let the RFModule know everything went well
	// so that it will then run the module
}

bool MVITEModule::interruptModule() {
	handlerPort.interrupt();
	ivel->stop();
	return true;
}

bool MVITEModule::close() {
	//close ports
	handlerPort.close();
	portOut.interrupt();
	portOut.close();

	/* stop the thread */
	dThread->stop();

	ivel->stop();
	return true;
}

bool MVITEModule::respond(const Bottle& command, Bottle& reply) {
	string helpMessage = string(getName().c_str()) + " commands are: \n"
			+ "help \n" + "quit \n" + "set thr <n> ... set the threshold \n"
			+ "(where <n> is an integer number) \n";

	reply.clear();

	if (command.get(0).asString() == "quit") {
		reply.addString("quitting");
		return false;
	} else if (command.get(0).asString() == "help") {
		cout << helpMessage;
		reply.addString("ok");
	} else if (command.get(0).asString() == "set") {
		if (command.get(1).asString() == "alpha") {
			alpha = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "beta") {
			beta = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "noise") {
			noise = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "target") {
			for (int i = 0; i < min(command.size() - 2, (int) newgoal.length());
					i++)
				newgoal[i] = command.get(2 + i).asDouble();
			/*bool ok = iencs->getEncoders(newgoal.data());
			 newgoal = DEG2RAD * newgoal;
			 for (int i=0; i<min(command.size()-2,(int)newgoal.length()); i++)	newgoal[i] = command.get(2+i).asDouble();*/
			reply.addString("ok");
			if (this->dThread->isStopping()) {
				//dThread->run();
				dThread->start();
				cout << "module is started" << endl;
			}

			newgoalreceived = true;
		} else if (command.get(1).asString() == "stop") {

			/* stop the thread */
			cout << "module is stopped" << endl;
			bool stopped = dThread->stop();
			ivel->stop();
			sleep(0.01);
			ivel->stop();
			if (stopped)
				reply.addString("ok");
			else
				reply.addString("could not stop, try again");
		} else {
			reply.addString("nothing to be set");
		}

	}
	return true;
}

bool MVITEModule::openPorts() {
	if (!portOut.open(outputPortName.c_str())) {
		cout << ": unable to open output port " << endl;
		return false; // unable to open; let RFModule know so that it won't run
	}

	return true;
}

bool MVITEModule::init() {
	bool ok = false;
	int axes;
	//bool ok = iencs->getAxes(&axes);
	while (!ok) {
		ok = iencs->getAxes(&axes);
		sleep(0.1);
	}
	if (!ok)
		return ok;

	dim = axes;
	newgoalreceived = false;
	newgoal = Vector(dim);
	x = Vector(dim);
	v = Vector(dim);
	xd = Vector(dim);
	xs = Vector(dim);

	//ok = iencs->getEncoders(x.data());
	ok = false;
	while (!ok) {
		ok = iencs->getEncoders(x.data());
		sleep(0.1);
	}
	if (!ok)
		return ok;
	x = DEG2RAD * x;
	v.zero();
	xd = x;
	xs = x;
	newgoal = x;
	this->armmovement = new Vite(xd, xs, alpha, beta, noise);
//	Vector tmp_xs1 = xs.subVector(0, 6); //Arm dimension
//	Vector tmp_xd1 = xd.subVector(0, 6);
//
//	Vector tmp_xs2 = xs.subVector(7, dim); //Hand dimension
//	Vector tmp_xd2 = xd.subVector(7, dim);
//	this->armmovement = new Vite(tmp_xd1, tmp_xs1, alpha, beta, noise);
//	this->handmovement = new Vite(tmp_xd2, tmp_xs2, alpha, 2*beta, noise);
	cout << x.toString() << endl;
	cout << "alpha = " << alpha << endl;
	cout << "beta = " << beta << endl;
	cout << "noise = " << noise << endl;

	return true;
}
bool MVITEModule::process() {
//	v.zero();
//	int steps = 0;
//	int maxSteps =3000;
//	//get feedback from Robot part
//	bool ok,ok2;
//	struct timespec delay;
//		delay.tv_sec = 0;
//		//delay.tv_nsec = 500000000L;  /* Half a second in nano's */
//		delay.tv_nsec = 10000000L;  /* a second in nano's */
//	while (maxSteps > steps && 0== yarp::math::norm(v)) {
//		//feedback
//		 ok = iencs->getEncoders(x.data());
//		 ok2 = iencs->getEncoderSpeeds(v.data());
//		if (!ok || !ok2) {
//			cout << "could not read encoders" << endl;
//			continue;
//		}
//
//		x = DEG2RAD * x;
//		v = DEG2RAD * v;
//		steps++;
//		nanosleep(&delay, NULL);
//	}
	//receive goal
	if (this->newgoalreceived) {
		cout << "[MVITE] setting new goal: ";
		this->newgoalreceived = false;
		//set new target

		xd = newgoal;
		//remember start
		xs = x;
		armmovement->setNewTask(xd, xs);
//		Vector tmp_xs1 = xs.subVector(0, 6); //Arm dimension
//		Vector tmp_xd1 = xd.subVector(0, 6);
//
//		Vector tmp_xs2 = xs.subVector(7, dim); //Hand dimension
//		Vector tmp_xd2 = xd.subVector(7, dim);
//		armmovement->setNewTask(tmp_xd1, tmp_xs1);
//		handmovement->setNewTask(tmp_xd2, tmp_xs2);
		cout << newgoal.toString() << endl;
		Vector xddeg = RAD2DEG * xd;
		ipos->positionMove(xddeg.data());

	}

	//calculate MVITE update step
//	Vector randomVec = yarp::math::Rand::vector(xd.length());
//	double normxdx = yarp::math::norm(xd - x);
//	Vector d = xd - xs;
//	Vector u = xd - x + noise * normxdx * randomVec;
//	double g = norm(d - u);
//	//Vector a = -alpha * v + beta * g * u;
//	Vector a = alpha * (-1 * v + beta * g * u);
//	v += a;
//	Vector tmp_x1 = x.subVector(0, 6); //Arm dimension
//	Vector tmp_v1 = v.subVector(0, 6);
//
//	Vector tmp_x2 = x.subVector(7, dim); //Hand dimension
//	Vector tmp_v2 = v.subVector(7, dim);
//	Vector v1 = this->armmovement->step(tmp_x1, tmp_v1);
//	Vector v2 = this->handmovement->step(tmp_x2, tmp_v2);
//	v.setSubvector(0, v1);
//	bool ok3 = v.setSubvector(v1.length(), v2);
//	if (!ok3){
//		cout<<"				could not set subvector to command "<<endl;
//	}
	//IF NO FDB or in position mode, integrate state

	v = this->armmovement->step(x, v);
	x += v;

	//publish new JAs
	//Vector vdeg = RAD2DEG * v;
	//ivel->velocityMove(vdeg.data());
	//Vector xddeg = RAD2DEG * xd;
	//ipos->positionMove(xddeg.data());

	//cout <<"[VITE] next Position   : "<< x.toString() << endl;
	//cout << "[VITE] target Position : " << xd.toString() << endl;
	//cout <<"[VITE] command velocity: "<< v.toString() << endl;

	//publish status
	double normv = norm(v);
	if (normv > 0.001) {
		//this->OUT_DS_STATUS->publish(DSStatus::CONVERGING());
	} else {
		//std::cout << "\t[MVITE::onProcess] converged" << std::endl;
		//this->OUT_DS_STATUS->publish(DSStatus::REACHED());
	}
	return true;

}
/* Called periodically every getPeriod() seconds */
bool MVITEModule::updateModule() {
	return true;
}

double MVITEModule::getPeriod() {
	/* module periodicity (seconds), called implicitly by myModule */
	return 0.1;
}

bool MVITEModule::openRobotInterface() {
	Property options;
	options.put("device", "remote_controlboard");
	std::string localportname = "/";
	localportname += getName();
	localportname += "/" + robotName + "/";
	localportname += bodyPart;
	options.put("local", localportname.c_str());
	std::string robotportname = "/";
	robotportname += robotName;
	robotportname += "/";
	robotportname += bodyPart;
	options.put("remote", robotportname.c_str());

	polydriver.open(options);
	polydriver.view(iencs);
	polydriver.view(ivel);
	polydriver.view(ipos);
	ivel->setVelocityMode();
	if ((!iencs) || (!ivel) || (!ipos)) {
		return false;
	} else {
		return true;
	}
}

