/**
 * @file MarkerTrackerModule.cpp
 * @brief implementation of the MarkerTrackerModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */

#include "iCub/RandomMovementModule.h"
#include "iCub/Vite.h"
#include <yarp/math/Math.h>
#include <yarp/math/Rand.h>
#include <iostream>
#include <string>
using namespace yarp::os;
using namespace yarp::sig;
using namespace yarp::dev;
using namespace yarp::math;

#define RAD2DEG    (180.0/M_PI)
#define DEG2RAD    (M_PI/180.0)
/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool RandomMovementModule::configure(yarp::os::ResourceFinder &rf) {

	/* Process all parameters from both command-line and .ini file */
	/* get the module name which will form the stem of all module port names */
	cout << "Found config file in: " << rf.getContextPath() << endl;
	moduleName = rf.check("name", Value("RandomHeadMovements"),
			"module name (string)").asString();

	inpPortName = "/";
	inpPortName += moduleName;
	inpPortName += rf.check("InputPort", Value("/command:i"),
			"module name (string)").asString();
	outPortName = "/";
	outPortName += moduleName;
	outPortName += rf.check("OutPort", Value("/status:o"),
			"module name (string)").asString();

	this->alpha = rf.check("alpha", Value(0.1)).asDouble();

	this->beta = rf.check("beta", Value(0.52)).asDouble();

	this->thresh = rf.check("thresh", Value(0.01)).asDouble();
	this->dim = rf.check("dim", Value(0)).asInt();
	cout << "Chosen parameters for Movement control: alpha " << this->alpha
			<< " beta " << this->beta << endl;
	/*
	 * before continuing, set the module name before getting any other parameters,
	 * specifically the port names which are dependent on the module name
	 */
	setName(moduleName.c_str());

	/* now, get the rest of the parameters */
	/*
	 * get the robot name which will form the stem of the robot ports names
	 * and append the specific part and device required
	 */
	robotName =
			rf.check("robot", Value("icub"), "Robot name (string)").asString();
	bodyPart =
			rf.check("part", Value("head"), "module name (string)").asString();
	cout << "connection to Robot name: " << robotName.c_str() << " BodyPart: "
			<< bodyPart.c_str() << endl;
	robotPortName = "/" + robotName + "/" + bodyPart;

	/* get the name of the input and output ports, automatically prefixing the module name by using getName() */
	/*  inputPortName         = "/";
	 inputPortName        += getName(
	 rf.check("myInputPort",
	 Value("/MarkerTrackerModule/image:i"),
	 "Input image port (string)").asString()
	 );*/

	/* do all initialization here */
	/*
	 * attach a port of the same name as the module (prefixed with a /) to the module
	 * so that messages received from the port are redirected to the respond method
	 */
	handlerPortName = "/";
	handlerPortName += getName(); // use getName() rather than a literal

	if (!handlerPort.open(handlerPortName.c_str())) {
		cout << getName() << ": Unable to open port " << handlerPortName
				<< endl;
		return false;
	}

	attach(handlerPort); // attach to port

	attachTerminal(); // attach to terminal
	startControl_b = false;
	setRandomTarget = false;
	/* opening ports */
	if (!portIn.open(inpPortName.c_str())) {
		std::cout << ": unable to open input port " << std::endl;
		return false; // unable to open; let RFModule know so that it won't rrun
	}

	if (!portOut.open(outPortName.c_str())) {
		cout << ": unable to open output port " << endl;
		return false; // unable to open; let RFModule know so that it won't run
	}

	/* create the thread and pass pointers to the module parameters */
	dThread = new RandomMovementThread(this);

	/* now start the thread to do the work */

	dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

	return true; // let the RFModule know everything went well
	// so that it will then run the module
}

bool RandomMovementModule::interruptModule() {

	return true;
}
void RandomMovementModule::stopMovement() {
	ivel->stop();
	sleep(0.1);
	ivel->stop();
}
void RandomMovementModule::process() {
	printf("random.");
	double dist = 1;
	bool test = false;
	/*** The main loop. It processes a frame per period ***/
	/* reads encoders */
	const Vector min(tmpMin);
	const Vector max(tmpMax);
	iencs->getEncoders(encoders.data());
	Vector tmp_xs = encoders.subVector(0, this->dim - 1) * DEG2RAD;
	Vector tmp_v = this->v.subVector(0, this->dim - 1) * DEG2RAD;
	int steps = 0;
	struct timespec delay;
	delay.tv_sec = 0;
	//delay.tv_nsec = 500000000L;  /* Half a second in nano's */
	delay.tv_nsec = 1000000L; /* a second in nano's */
	while (isStopping() != true) { // the thread continues to run until isStopping() returns true

		// DO control if startContorl_b is set
		while (dist > this->thresh && !isStopping() && startControl_b) {
			iencs->getEncoders(encoders.data());
			steps = 0;
			//get feedback from Robot part
			while (maxSteps > steps
					&& 0
							== yarp::math::norm(
									encoders.subVector(0, this->dim - 1)
											* DEG2RAD - (tmp_xs))) {
				iencs->getEncoders(encoders.data());
				steps++;
				nanosleep(&delay, NULL);
			}
			//debug output
			cout << "waited for encoders: " << steps << endl;
			//iencs->getEncoderSpeeds(v.data());
			cout << "current position [rad]: " << tmp_xs.toString(2, 0) << endl;
			//cout<< "target position [rad]:  " << vite->xd.toString(2,0) <<endl;

			tmp_v = (encoders.subVector(0, this->dim - 1) * DEG2RAD) - tmp_xs; //v.subVector(0,2)*DEG2RAD;
			tmp_xs = encoders.subVector(0, this->dim - 1) * DEG2RAD;
			dist = yarp::math::norm(targets - (tmp_xs));
			tmp_v = vite->step(tmp_xs, tmp_v);

			cout << "target  vel in [rad/s] " << tmp_v.toString(2, 1) << endl;
			cout << "current vel in [rad/s] " << v.toString(2, 1)
					<< " Dist to targ: " << dist << endl;
			cout << "current pos in   [deg] " << encoders.toString(2, 1)
					<< " Dist to targ: " << dist << endl;
			v.zero();
			v.setSubvector(0, tmp_v * RAD2DEG);
			cout << "sending velocity command now " << v.toString(2, 0) << endl;
			//COMMAND to ROBOT NEEDS TO BE IN [deg/s]
			if (strcmp(this->bodyPart.c_str(), "right_arm") == 0) {
				//test = ivel->velocityMove(v.subVector(0,6).data());
				for (int i = 0; i < 7; i++)
					ivel->velocityMove(i, v[i]);
				for (int i = 7; i < 16; i++)
					ipos->positionMove(i, vite->xd[i] * RAD2DEG);
			} else {
				test = ivel->velocityMove(v.data());
				//for (int i=0; i<this->dim; i++)	ipos->positionMove(i, vite->xd[i]*RAD2DEG);
			}

			listenToCommand();
		}
		dist = 1;
		Vector wtmp = this->targets;
		bool notChanged = true;
		if (setRandomTarget && startControl_b) {
			//while (0.5<yarp::math::norm(targets -(wtmp))){
			//TODO needs to be adapted depending on body part
			//currently it works only for head control
			this->targets = yarp::math::Rand::vector(min, max);
			//	}
			vite->setNewTask(this->targets, tmp_xs);
			notChanged = false;
		}

		/*ivel->stop();
		 sleep(0.1);
		 ivel->stop();*/

		while (!setRandomTarget && notChanged) {
			//ivel->stop();
			listenToCommand(true);
			if (0 < yarp::math::norm(targets - (wtmp))) {
				notChanged = false;
			} else {
				vite->setNewTask(this->targets, tmp_xs);
			}
			nanosleep(&delay, NULL);
		}

		listenToCommand();

		/* print something at each iteration so we see when the thread runs */
		//printf(".");printf("RandomMovementThread:run\n");
	}

}
bool RandomMovementModule::init() {
	tmpMax = Vector(3);
	tmpMin = Vector(3);
	tmpMax[0] = 10 * DEG2RAD;
	tmpMax[1] = 0 * DEG2RAD;
	tmpMax[2] = 20 * DEG2RAD;
	tmpMin[0] = 0 * DEG2RAD;
	tmpMin[1] = 0 * DEG2RAD;
	tmpMin[2] = -20 * DEG2RAD;
	const Vector min(tmpMin);
	const Vector max(tmpMax);
	this->targets = yarp::math::Rand::vector(min, max);
	string robotPortName = "/" + robotName + "/" + this->bodyPart;
	/* initialize here variables */
	printf("RandomMovementThread:starting\n");

	string localPortName = "/local/" + this->bodyPart;
	Property options;
	options.put("device", "remote_controlboard");
	options.put("local", localPortName.c_str());

	/* substitute icubSim with icub for use with the real robot */
	options.put("remote", robotPortName.c_str());

	dd.open(options);
	dd.view(iencs);
	dd.view(ivel);
	dd.view(ipos);

	if ((!iencs) || (!ivel))
		return false;

	int joints;
	Vector tmp;

	iencs->getAxes(&joints);
	encoders.resize(joints);
	commands.resize(joints);
	tmp.resize(joints);
	/* we need to set reference accelerations used to generate the velocity */
	/* profile, here 50 degrees/sec^2 */

	int i;
	for (i = 0; i < joints; i++) {
		commands[i] = 50.0;
	}
	cout << "size detected " << joints << " dimension in config file: "
			<< this->dim << endl;
	ivel->setRefAccelerations(commands.data());

	for (i = 0; i < joints; i++) {
		tmp[i] = 40.0;
	}
	ipos->setRefSpeeds(tmp.data());

	iencs->getEncoders(encoders.data());
	Vector tmp_xs = encoders.subVector(0, this->dim - 1);
	//Vite vite = Vite(this->targets , tmp_xs,0.1, 0.32,0.1);
	targets.resize(this->dim);
	this->vite = new Vite(tmp_xs, tmp_xs, this->alpha, this->beta, 0.01);
	this->v = Vector(encoders.length());

	v.zero();

	count = 0;
	return true;
}

bool RandomMovementModule::close() {
	ivel->stop();

	dd.close();
	handlerPort.close();
	portIn.close();
	portOut.close();
	/* stop the thread */
	dThread->stop();
	printf("Calling RandomMovementModule::close()");

	return true;
}

bool RandomMovementModule::respond(const Bottle& command, Bottle& reply) {
	string helpMessage = string(getName().c_str()) + " commands are: \n"
			+ "help \n" + "quit \n";

	reply.clear();

	if (command.get(0).asString() == "quit") {
		reply.addString("quitting");
		return false;
	} else if (command.get(0).asString() == "help") {
		cout << helpMessage;
		reply.addString("ok");
	}

	return true;
}

/* Called periodically every getPeriod() seconds */
bool RandomMovementModule::updateModule() {

	return true;
}

double RandomMovementModule::getPeriod() {
	/* module periodicity (seconds), called implicitly by myModule */
	return 0.1;
}
void RandomMovementModule::startControl() {
	startControl_b = true;
	iencs->getEncoders(encoders.data());
	Vector tmp_xs = encoders.subVector(0, this->dim - 1) * DEG2RAD;
	this->vite->setNewTask(targets, tmp_xs);
}

void RandomMovementModule::listenToCommand(bool blocking) {
	if (blocking) {
		cout << "blocking call " << endl;
	}
	Bottle* input = portIn.read(blocking);
	if (input != NULL) {
		cout << "received " << input->toString().c_str() << endl;
		Bottle& reply = portIn.prepare();
		reply.clear();

		if (0 == strcmp(input->get(0).asString(), "set")) {
			reply.addString("[OK]");
			if (0 == strcmp(input->get(1).asString(), "start")) {
				reply.addString("start");
				setRandomTarget = false;
				if (0 == strcmp(input->get(2).asString(), "random")) {
					reply.addString("random");
					setRandomTarget = true;
					cout << "starting control random movements " << endl;
				} else {
					cout << "starting control " << endl;
				}

				startControl();
			}
			if (0 == strcmp(input->get(1).asString(), "alpha")) {
				reply.addString("alpha");
				vite->setAlpha(input->get(2).asDouble());
				this->alpha = input->get(2).asDouble();
			}
			if (0 == strcmp(input->get(1).asString(), "beta")) {
				reply.addString("beta");
				vite->setBeta(input->get(2).asDouble());
				this->beta = input->get(2).asDouble();
			}

			if (0 == strcmp(input->get(1).asString(), "stop")) {
				reply.addString("stop");
				stopMovement();
				startControl_b = false;
			}

			if (0 == strcmp(input->get(1).asString(), "target")) {
				reply.addString("target");

				cout << "old targets: " << targets.toString(2, 0) << endl;
				targets.zero();
				for (int i = 0; i < targets.length(); i++) {
					targets[i] = input->get(2 + i).asDouble();
				}
				if (input->size() > targets.length() + 2) {
					if (0
							== strcmp(
									input->get(targets.length() + 3).asString(),
									"alpha")) {
						reply.addString("alpha");
						this->alpha =
								input->get(targets.length() + 4).asDouble();
						vite->setAlpha(this->alpha);
					}
				}
				setRandomTarget = false;
				startControl();
				cout << "targets received: " << targets.toString(2, 0) << endl;
			}

		} else if (0 == strcmp(input->get(0).asString(), "get")) {
			if (0 == strcmp(input->get(1).asString(), "alpha")) {
				reply.addDouble(alpha);
			} else if (0 == strcmp(input->get(1).asString(), "beta")) {
				reply.addDouble(beta);
			}
		} else {
			reply.addString("[NOTOK]");
		}
		portIn.write(false);
		sentStatus();
	}

}
void RandomMovementModule::sentStatus() {
	//Bottle output;
	Bottle& output = portOut.prepare();
	output.clear();

	output.addString("alpha_beta");
	output.addDouble(alpha);
	output.addDouble(beta);
	output.addString("velocity");
	for (int i = 0; i < v.length(); i++) {
		output.addDouble(v[i]);
	}
	//cout << "writing " << output.toString().c_str() << endl;
	portOut.write();
}

