classdef ELM < handle
    %ELM Extreme Learning Machine
    %   Single layer feed-forward network
    
    properties
        inpDim = 0; %dimension of input
        outDim = 0; %dimension of output
        out = [];

        hidDim = 10;    %number of neurons in the hidden layer
        Winp = [];      %weights from input layer to hidden layer
        Wout = [];      %weights from hidden layer to output layer
        a = [];         %slope parameters of activation functions
        b = [];         %bias parameters of activation functions
        
        reg = 1e-6;     %regularization parameter for ridge regression
        BIP = 1;        %flag for using Batch Intrinsic Plasticity (BIP)
        mu = 0.2;       %desired mean activity parameter for BIP

        batchSize = 1000; %number of samples used for a minibatch
    end
    
    methods
        function l = ELM(inpDim, outDim, spec)
            l.inpDim = inpDim;
            l.outDim = outDim;
            l.out = zeros(1,l.outDim);
            
            if nargin > 2
                for s=1:size(spec,1)
                    try
                        l.(spec{s,1}) = spec{s,2};
                    catch me
                        disp(me.message);
                    end
                end
            end
        end

        function init(l, X)
            l.a = ones(1,l.hidDim);
            l.b = 2 * rand(1,l.hidDim) - ones(1,l.hidDim);
            l.Winp = 2 * rand(l.hidDim,l.inpDim) - ones(l.hidDim,l.inpDim);
            l.Wout = 2 * rand(l.outDim,l.hidDim) - ones(l.outDim,l.hidDim);
            
            if nargin > 1 && l.BIP
                l.bip(X);
            end
        end
        
        function train(l, X, Y)
            if size(X,1) <= l.batchSize
                hs = l.calcHiddenStates(X);
                l.Wout = (hs'*hs + l.reg * eye(l.hidDim))\(hs' * Y);
            else
                HTH = zeros(l.hidDim, l.hidDim);
                HTY = zeros(l.hidDim, l.outDim);                
                numBatches = floor(size(X,1)/l.batchSize);
                rest = mod(size(X,1),l.batchSize);                
                for b=1:numBatches
                    H = l.calcHiddenStates(X((b-1)*l.batchSize+1:b*l.batchSize,:));
                    HTH = HTH + H' * H;
                    HTY = HTY + H' * Y((b-1)*l.batchSize+1:b*l.batchSize,:);
                end
                if rest > 0
                    H = l.calcHiddenStates(X(numBatches*l.batchSize+1:end,:));
                    HTH = HTH + H' * H;
                    HTY = HTY + H' * Y(numBatches*l.batchSize+1:end,:);
                end                
                HTH = HTH + diag(repmat(l.reg, 1, l.hidDim));
                l.Wout = HTH\HTY;
            end
        end
        
        function [Y H] = apply(l, X)
            if size(X,1) <= l.batchSize
                H = l.calcHiddenStates(X);
                Y = H * l.Wout;
            else
                Y = zeros(size(X,1), l.outDim);
                numBatches = floor(size(X,1)/l.batchSize);
                rest = mod(size(X,1),l.batchSize);
                for b=1:numBatches
                    H = l.calcHiddenStates(X((b-1)*l.batchSize+1:b*l.batchSize,:));
                    Y((b-1)*l.batchSize+1:b*l.batchSize,:) = H * l.Wout;
                end
                if rest > 0
                    H = l.calcHiddenStates(X(numBatches*l.batchSize+1:end,:));
                    Y(numBatches*l.batchSize+1:end,:) = H * l.Wout;
                end
            end
            l.out = Y(end,:);
        end
        
        function [H G] = calcHiddenStates(l, X)
            numSamples = size(X,1);
            atemp = repmat(l.a, numSamples, 1);
            btemp = repmat(l.b, numSamples, 1);
            G = X * l.Winp';
            H = 1./(1+exp(-atemp .* G - btemp));
            %H = tanh(G + btemp);
        end
        
        function bip(l, X)
            numSamples = size(X,1);
            
            PD = ProbDistUnivParam('exponential', l.mu);
            G = X * l.Winp';

            for hn = 1:l.hidDim
                targets = random(PD, 1, numSamples);

                hightars = length([targets(targets>=1),targets(targets<=0)]);
                while hightars > 0
                    further_targets = random(PD, 1, hightars);
                    targets = [targets(targets<1&targets>0),further_targets];
                    hightars = length([targets(targets>=1),targets(targets<=0)]);
                end
                targets = sort(targets);
                
                s = sort(G(:,hn));
                Phi = [s,ones(numSamples,1)];
                targetsinv = -log(1./targets - 1); %apply inverse activation function
                
                w = pinv(Phi)*targetsinv';                
                
                l.a(hn) = w(1);
                l.b(hn) = w(2);
            end
        end
        
        function new = copy(this)
            new = feval(class(this));
            p = properties(this);
            for i = 1:length(p)
                if( isa(this.(p{i}), 'handle'))
                    new.(p{i}) = this.(p{i}).copy();
                elseif isa(this.(p{i}),'cell')
                    new.(p{i}) = deepCopyCellArray(numel(this.(p{i})));
                else
                    new.(p{i}) = this.(p{i});
                end
            end
        end
        
    end
    
end

