/**
 * @file Ropsci.cpp
 * @brief implementation of the RopsciModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */ 

#include "iCub/RopsciModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool RopsciModule::configure(yarp::os::ResourceFinder &rf) {    
    /*
     * PLEASE remove useless comments when writing actual code. If needed then use Doxygen comments and tags.
     */

    /* Process all parameters from both command-line and .ini file */
    /* get the module name which will form the stem of all module port names */
    moduleName            = rf.check("name", 
                           Value("Ropsci"), 
                           "module name (string)").asString();

    /*
    * before continuing, set the module name before getting any other parameters, 
    * specifically the port names which are dependent on the module name
    */
    setName(moduleName.c_str());

    
    /* get the name of the input and output ports, automatically prefixing the module name by using getName() */
    inputPortName         = "/";
    inputPortName        += getName(
                           rf.check("myInputPort", 
                           Value("/image:i"),
                           "Input image port (string)").asString()
                           );

    outputPortName        = "/";
    outputPortName       += getName(
                           rf.check("myOutputPort", 
                           Value("/rps:o"),
                           "Class label <rock, paper, scissor> output port (string)").asString()
                           );

	record = false;
	saveto = rf.check("saveto", Value("./rps"), "Key value (string)").asString();
	imageidx = 0;


    /* do all initialization here */
    /*
    * attach a port of the same name as the module (prefixed with a /) to the module
    * so that messages received from the port are redirected to the respond method
    */
    handlerPortName =  "/";
    handlerPortName += getName();         // use getName() rather than a literal 
	handlerPortName += "/command";
    if (!handlerPort.open(handlerPortName.c_str())) {           
        cout << getName() << ": Unable to open port " << handlerPortName << endl;  
        return false;
    }
    attach(handlerPort); // attach to port
    attachTerminal(); // attach to terminal

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;  
		return false;
	}

	//init module variables
	this->init();

    /* create the thread and pass pointers to the module parameters */
    dThread = new RopsciThread(this);

    /* now start the thread to do the work */    
    dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

    return true ;     // let the RFModule know everything went well
                      // so that it will then run the module
}

bool RopsciModule::interruptModule() {
    handlerPort.interrupt();
    return true;
}

bool RopsciModule::close() {
    //close ports
    handlerPort.close();
    imagePortIn.interrupt();
    portOut.interrupt();
    imagePortIn.close();
    portOut.close();

	cvDestroyWindow("Ropsci");
	delete this->classifier;

    /* stop the thread */
    dThread->stop();

    return true;
}

bool RopsciModule::respond(const Bottle& command, Bottle& reply) {
  string helpMessage =  string(getName().c_str()) + 
                        " commands are: \n" +  
                        "help \n" + 
                        "quit \n" + 
                        "set thr <n> ... set the threshold \n" + 
                        "(where <n> is an integer number) \n";

  reply.clear(); 

  if (command.get(0).asString()=="quit") {
       reply.addString("quitting");
       return false;     
   }
   else if (command.get(0).asString()=="help") {
      cout << helpMessage;
      reply.addString("ok");
   }
   else if (command.get(0).asString()=="set") {
      if (command.get(1).asString()=="imageidx") {
         imageidx = command.get(2).asInt(); // set parameter value
         reply.addString("ok");
      } else if (command.get(1).asString()=="saveto") {
         saveto = command.get(2).asString(); // set parameter value
		 imageidx = 0;
         reply.addString("ok");
      }else if (command.get(1).asString()=="record") {
         record = (bool) command.get(2).asInt(); // set parameter value
         reply.addString("ok");
      }
   }
   return true;
}

bool RopsciModule::openPorts() {
    if (!imagePortIn.open(inputPortName.c_str()) ) {
        std::cout << ": unable to open input port " << std::endl;
        return false;  // unable to open; let RFModule know so that it won't rrun
    }

    if (!portOut.open(outputPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }

	return true;
}

/* Called periodically every getPeriod() seconds */
bool RopsciModule::updateModule() {

    return true;
}

double RopsciModule::getPeriod() {
    /* module periodicity (seconds), called implicitly by myModule */    
    return 0.1;
}

void RopsciModule::init() {
	img = 0;
	grayImg = 0;
	colorMaskImg = 0;
	contourImg = 0;
	roiImg = 0;
	gaborImg = 0;

	red_min = 95;
	green_min = 40;
	blue_min = 20;
	max_min_span = 0;
	min_value = 3;

	//load classifier from file
	nemo::RealMatrix winp = nemoReadRealMatrix("ELM/ELM_aWinp.mat");
	nemo::RealMatrix wout = nemoReadRealMatrix("ELM/ELM_Wout.mat");//.transpose();
	unsigned int inpDim = winp.cols();
	unsigned int hidDim = winp.rows();
	unsigned int outDim = wout.rows();

	//create ELM
	this->classifier = new amarsi::ELMLearner(inpDim, hidDim, outDim);
	this->linp = nemo::RealVector(nemo::dim(inpDim));
	this->lout = nemo::RealVector(nemo::dim(outDim));
	this->classifier->initialize(this->linp);
	//set params
	((amarsi::ELMLearner*) this->classifier)->setBias(nemoReadRealVector("ELM/ELM_b.vec"));
	((amarsi::ELMLearner*) this->classifier)->setWin(winp);
	((amarsi::ELMLearner*) this->classifier)->setWout(wout);
	this->initGaborFilters();


	// create a window
	//cvNamedWindow("Ropsci ", CV_WINDOW_AUTOSIZE); 
	//cvMoveWindow("Ropsci", 100, 100);
	//cvNamedWindow("Ropsci", CV_WINDOW_NORMAL);
	//cvResizeWindow("Ropsci", 640, 480);
	//cvMoveWindow("Ropsci", 0, 0);
}









//from nao_color_rule.cpp
void RopsciModule::skinColorFilter() {
	//cout << "create helpers..." << endl;
	CvSize size = cvSize(img->width, img->height);
	IplImage* red_channel = cvCreateImage(size, IPL_DEPTH_8U, 1);
	IplImage* green_channel = cvCreateImage(size, IPL_DEPTH_8U, 1);
	IplImage* blue_channel = cvCreateImage(size, IPL_DEPTH_8U, 1);
	this->colorMaskImg = cvCreateImage(size, IPL_DEPTH_8U, 1);

	unsigned char* red_data = (unsigned char*) red_channel->imageData;
	unsigned char* green_data = (unsigned char*) green_channel->imageData;
	unsigned char* blue_data = (unsigned char*) blue_channel->imageData;
	unsigned char* dst = (unsigned char*) this->colorMaskImg->imageData;
	//std::cout << "split channels..." << endl;
	cvSplit(img, blue_channel, green_channel, red_channel, NULL);

	//std::cout << "skin color detection..." << endl;
	const unsigned int length = this->colorMaskImg->width * this->colorMaskImg->height;
	ColorRule cr(red_min, green_min, blue_min, max_min_span, min_value);
	for (unsigned int i = 0; i < length; i++) {
		const unsigned char red(red_data[i]), green(green_data[i]), blue(blue_data[i]);
		dst[i] = cr.match(red, green, blue, max(red, max(green, blue)), min(red, min(green, blue))) << 7;
	}

	/* eliminate all regions within the
	 skin color mask that are smaller than 80%
	 of the largest region. */
	//cout << "eliminating regions..." << endl;
	this->eliminateRegions(colorMaskImg, 0.8);
	
	/*cvNamedWindow("skin color mask", CV_WINDOW_AUTOSIZE);
	cvShowImage("skin color mask", colorMaskImg);
	cvWaitKey(200);
	cvDestroyWindow("skin color mask");*/

	cvReleaseImage(&blue_channel);
	cvReleaseImage(&red_channel);
	cvReleaseImage(&green_channel);
}

bool RopsciModule::boundingBox() {
	//std::cout << "Boxing image..." << endl;
	roiImg = cvCreateImage(
			cvSize(colorMaskImg->width, colorMaskImg->height), IPL_DEPTH_8U,
			colorMaskImg->nChannels);

	/* Copy the original image as OpenCVs contour finder will
	 overwrite this image */
	contourImg = cvCloneImage(colorMaskImg);
	CvMemStorage* store = cvCreateMemStorage(0);
	CvSeq* contours;
	int numContours;

	// Find all contours within the skin color mask
	numContours = cvFindContours(contourImg, store, &contours);

	// No contours = no skin color
	if (numContours == 0) {
		cout << "No skin contours found\n" << endl;
		return false;
	}

	CvBox2D box;
	CvSeq* contUnified = contours;
	contours = contours->h_next;

	// Combine all the contours into one large set of contour pixels
	for (int i = 1; i < numContours; i++) {
		cvSeqInsertSlice(contUnified, 0, contours);
		contours = contours->h_next;
	}

	CvPoint2D32f bp[4];

	// Put a minimum-area bounding box around the contour pixels
	box = cvMinAreaRect2(contUnified);

	//check box size
	//cout << box.size.width << " " << box.size.height << endl;

	// Extract the corner points from the bounding box
	cvBoxPoints(box, bp);

	float max_length = -1.;
	int best_index = -1;
	float best_vector[2] = { 0, 0 };

	/* As the angle returned by the bounding box function has
	 proven to be unreliable at times, we calculate our own angle.
	 First, we find the longest side of the box. We have to
	 check all 4 edges of the bounding box as the ordering
	 of the edges may change*/
	for (int i = 0; i < 4; i++) {
		int j = (i + 1) % 4;
		float len = sqrt(
				pow((float) (bp[i].x - bp[j].x), 2.0f)
						+ pow((float) (bp[i].y - bp[j].y), 2.0f));
		//cout << "Len: " << len << " @ bp[i] = " << bp[i].x << " x " << bp[i].y << endl;
		if (max_length < len) {
			best_index = i;
			max_length = len;
			best_vector[0] = ((float) (bp[i].x - bp[j].x)) / len;
			best_vector[1] = ((float) (bp[i].y - bp[j].y)) / len;
		}
	}

	/* Eliminate all pixels at the border of the image.
	 This is necessary as the rotation used below will
	 use these pixels to extrapolate pixels in the rotated
	 image that were beyond the border in the original image,
	 and we want to avoid adding foreground pixels that aren't
	 actually there */
	for (int i = 0; i < colorMaskImg->height; ++i) {
		GETAT(colorMaskImg, i, 0) = 0;
		GETAT(colorMaskImg, i, colorMaskImg->width - 1) = 0;
	}

	for (int i = 0; i < colorMaskImg->width; ++i) {
		GETAT(colorMaskImg, 0, i) = 0;
		GETAT(colorMaskImg, colorMaskImg->height - 1, i) = 0;
	}

	/* Apply the binary skin color mask to the camera image */
	if (grayImg != NULL) {
		for (int i = 0; i < colorMaskImg->height; ++i) {
			for (int j = 0; j < colorMaskImg->width; ++j) {
				//cout << "GETAT(IplSkin_Img, " << i << ", " << j << ") = " << ((int) GETAT(IplSkin_Img, i, j)) << endl;
				if (GETAT(colorMaskImg, i, j) != 0) {
					GETAT(colorMaskImg, i, j) = GETAT(grayImg, i, j);
				}
			}
		}
	}

	/* Do foreground-only grayscale histogram normalization
	 for the resulting image */
	//equalizeHistoNoBG(IplSkin_Img);
	//cout << "Bounding box size: " << box.size.width << " x " << box.size.height << endl;
	/* Determine the scaling factor */
	float scale_width = MIN(box.size.width, colorMaskImg->width)
			/ (float) colorMaskImg->width;
	float scale_height = MIN(box.size.height, colorMaskImg->height)
			/ (float) colorMaskImg->height;
	/* We use the minimum scaling (maximum downscale factor) of the two axes
	 so that no pixels are lost during the scaling */
	float scale_factor = MAX(scale_width, scale_height);
	float angle = atan(best_vector[1] / best_vector[0]);

	//cout << "SCALE FACTOR: " << scale_factor << endl;

	/* Build the transformation matrix using the angle, scale factor and
	 bounding box center */
	float m[6];
	CvMat mat = cvMat(2, 3, CV_32F, m);

	m[0] = scale_factor * (float) cos(-angle);
	m[1] = scale_factor * (float) sin(-angle);
	m[3] = -m[1];
	m[4] = m[0];
	m[2] = box.center.x;
	m[5] = box.center.y;

	/* apply the transformation matrix. This does the rotation, scaling
	 and recentering all in one go */
	//cout << "Applying transformation matrix..." << endl;
	cvGetQuadrangleSubPix(colorMaskImg, roiImg, &mat);

	/* Release the images */
	cvReleaseMemStorage(&store);
	cvReleaseImage(&contourImg);
	//cvReleaseImage(&colorMaskImg);
	cvReleaseImage(&grayImg);

	return true;
}

void RopsciModule::gaborFilter() {
	//cout << "Gabor filtering image..." << endl;

	// Calculate the size of each 3x3 grid cell
	const int m = roiImg->height / 3;
	const int n = roiImg->width / 3;
	unsigned int i, j, k;

	// Initialize the gabor result image for the first time
	if (gaborImg == NULL) {
		gaborImg = cvCreateImage(
				cvSize(roiImg->width, roiImg->height),
				IPL_DEPTH_8U, 1);
	}

	/* Display the processed image. */
	//preview->render(roiImg, IW_BW);
	// For each of the 4 directions...
	int cnt = 0;
	double sumfeature = 0.0;
	for (k = 0; k < 4; k++) {
		// Build the gabor filter intensity image
		this->gaborFilters[k]->conv_img(roiImg, gaborImg, CV_GABOR_MAG);

		// Legacy loop, will be optimized out by the compiler
		for (i = 0; i < 2; i++) {
			// For each of the 3 cells
			for (j = 0; j < 3; j++) {
				// Set the image region of interest to the current cell
				cvSetImageROI(gaborImg, cvRect(i * n, j * m, n, m));

				// The feature for that cell will be the sum of all pixel
				// intensities within that cell
				const float feature = cvSum(gaborImg).val[0];
				sumfeature += feature;
				//cout << feature << " ";

				// Add the feature to the new feature vector
				//currentData.push_back(feature);
				linp[cnt++] = feature;

				// Draw the feature vectors / gabor response magnitudes (green lines)
				//const int centerx = (i * n) + (n / 2);
				//const int centery = (j * m) + (m / 2);
				//const int length = (int) ((feature / 3000000.0) * (m / 2));
				//const float angle = (PI / 4) * k;
				//const int dx = (cosf(angle) * length);
				//const int dy = (sinf(angle) * length);
				//preview->render_line(centerx, centery, centerx - dx, centery - dy, 0, 255, 0);

				//cout << "Feature @ " << i << " / " << j << " = " << features[i * 3 + j].val[0] << endl;
			}
		}
	}
	//cout << endl;
	linp = (1.0/sumfeature) * linp;

	/* Display the data on the screen. */
	//preview->draw ();

	if (record) {
		char name[100];
		sprintf(name,"%s_roi_%03d.pgm", saveto.c_str(), imageidx);
		cout << "writing image and features " << name;
		cvSaveImage(name,roiImg);
		sprintf(name,"%s_inp_%03d.vec", saveto.c_str(), imageidx++);
		nemoWriteRealVector(name, linp);
		cout << " done!" << endl;
		//sleep(0.1);
	}

	//cvNamedWindow("gabor filtered img", CV_WINDOW_AUTOSIZE);
	//cvShowImage("gabor filtered img", gaborImg);
	//cvWaitKey(10);
	//cvDestroyWindow("gabor filtered img");

	cvReleaseImage(&roiImg);
}

void RopsciModule::convertToGreyScale(IplImage* IplAktuellesBild) {
	// Initialize storage space only during the first run
	grayImg = cvCreateImage(
				cvSize(IplAktuellesBild->width, IplAktuellesBild->height),
				IPL_DEPTH_8U, 1);
	cvCvtColor(IplAktuellesBild, grayImg, CV_BGR2GRAY);
}

void RopsciModule::initGaborFilters() {
	const double Sigma = 2 * M_PI;
	const double F = sqrt(2.0);

	// Create the gabor kernels in 4 different directions
	for (int i = 0; i < 4; i++) {
		gaborFilters[i] = new CvGabor(2 * i, 3, Sigma, F);
	}
}

/**
 * Update the Region Map
 */
inline void RopsciModule::updateRegionMap(std::vector<uint>* const regionMap, const uint r1, const uint r2) {
	assert(r1 != r2);

	uint trueR1 = r1, trueR2 = r2;

	/* If we're dealing with a region chain, find the
	 final region in the chain */
	while ((*regionMap)[trueR1] != trueR1) {
		trueR1 = (*regionMap)[trueR1];
	}

	while ((*regionMap)[trueR2] != trueR2) {
		trueR2 = (*regionMap)[trueR2];
	}

	//assert(trueR1 != trueR2);
	if (trueR1 == trueR2) {
		return;
	}

	/* Map the larger region to the base region of the smaller one */
	if (trueR1 < trueR2) {
		(*regionMap)[trueR2] = trueR1;
	} else {
		(*regionMap)[trueR1] = trueR2;
	}
}

/**
 * Eliminate all regions that are smaller than the size of the
 * largest region multiplied with a threshold
 */
void RopsciModule::eliminateRegions(IplImage* sourceImg, float threshold) {
	const int m = sourceImg->height;
	const int n = sourceImg->width;
	int i, j;
	uint k;

	/* The first region will be region 1 as region 0 a dummy region */
	uint regionCounter = 1;
	uint tmpRegion, neighbourRegion;
	int pixelValue;

	uint largestRegionSize = 0;

	std::vector<uint> regionMap;
	std::vector<uint> regionSizes;
	std::vector<uint> nI, nJ;

	Eigen::MatrixXf regions = Eigen::MatrixXf::Zero(m, n);

	/* The dummy region 0 has a size of 0 */
	regionSizes.clear();
	regionSizes.push_back(0);

	/* Dummy mapping for the dummy region */
	regionMap.clear();
	regionMap.push_back(0);

	/* Iteratre over each pixel */
	for (i = 0; i < m; ++i) {
		for (j = 0; j < n; ++j) {
			tmpRegion = 0;
			pixelValue = GETAT(sourceImg, i, j);

			/* Ignore background pixels in region building */
			if (pixelValue == 0) {
				continue;
			}

			/* Add all valid neighbour pixel coordinates to nI / nJ
			 nI / nJ will contain the absolute coordinates for valid
			 neighbours of the current pixel. */
			nI.clear();
			nJ.clear();

			if (j > 0) {
				/* The pixel to the left */
				nI.push_back(i);
				nJ.push_back(j - 1);

				if (i > 0) {
					/* Pixel to the upper left */
					nI.push_back(i - 1);
					nJ.push_back(j - 1);
				}
			}

			if (i > 0) {
				/* Pixel to the top */
				nI.push_back(i - 1);
				nJ.push_back(j);
				if (j < (sourceImg->width - 1)) {
					/* Pixel to the upper right */
					nI.push_back(i - 1);
					nJ.push_back(j + 1);
				}
			}
			/* For each valid neighbour... */
			for (k = 0; k < nI.size(); ++k) {
				if (pixelValue == GETAT(sourceImg, nI[k], nJ[k])) {
					neighbourRegion = regions(nI[k], nJ[k]);
					/* If the pixel values match */
					if ((tmpRegion > 0) && (tmpRegion != neighbourRegion)) {
						/* If we have already found a different region for this pixel,
						 then mark this region and the new one as equal */
						updateRegionMap(&regionMap, tmpRegion, neighbourRegion);
					} else {
						/* Add this pixel to the existing region */
						tmpRegion = neighbourRegion;
					}
				}
			}

			if (tmpRegion == 0) {
				/* Create a new region for this pixel */
				tmpRegion = regionCounter++;
				regionMap.push_back(tmpRegion);
				regionSizes.push_back(0);
			}
			/* Add this pixel to the region */

			regions(i, j) = tmpRegion;
			regionSizes[tmpRegion]++;
		}
	}

	/* Convert all "chains" within the region map / region tree to
	 direct references */
	for (k = 1; k < regionMap.size(); k++) {
		tmpRegion = k;
		while (regionMap[tmpRegion] != tmpRegion) {
			tmpRegion = regionMap[tmpRegion];
		}
		regionMap[k] = tmpRegion;
	}

	/* Calculate combined region sizes */
	for (k = 1; k < regionCounter; k++) {
		if (regionMap[k] != k) {
			regionSizes[regionMap[k]] += regionSizes[k];
			regionSizes[k] = 0;
		}
	}

	/* Find the largest region */
	for (k = 1; k < regionCounter; k++) {
		if (regionSizes[k] > largestRegionSize) {
			largestRegionSize = regionSizes[k];
		}
	}

	/* Eliminate small regions */
	for (i = 0; i < m; ++i) {
		for (j = 0; j < n; ++j) {
			/* Ignore background pixels */
			if (GETAT(sourceImg, i, j) == 0) {
				continue;
			}

			/* Look up the base region that this region belongs to */
			tmpRegion = regionMap[regions(i, j)];

			/* Delete all pixels where the base region is smaller than the
			 threshold*/
			if (regionSizes[tmpRegion]
					<= (uint) (largestRegionSize * threshold)) {
				//cout << "Eliminating Pixel at " << i << " / " << j << endl;
				GETAT(sourceImg, i, j) = 0;
			}
		}
	}
}


int RopsciModule::classify() {
	//cout << linp << endl;
	this->lout = this->classifier->execute(this->linp);
	cout << lout << endl;
	nemo::RealVector::Iterator itor = std::max_element(lout.begin(), lout.end());
	int winner = std::distance(lout.begin(), itor);
	//if (*itor < 0.0)	winner = -1;
	switch (winner) {
		case Ropsci_Rock: cout << "rock detected" << endl;	break;
		case Ropsci_Paper: cout << "paper detected" << endl;	break;
		case Ropsci_Scissor: cout << "scissor detected" << endl;	break;
		default: cout << "none" << endl;
	}
	return winner;
}







