/**
 * @file RopsciSequencer.cpp
 * @brief implementation of the RopsciSequencerModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */ 

#include "iCub/RopsciSequencerModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool RopsciSequencerModule::configure(yarp::os::ResourceFinder &rf) {    
    /*
     * PLEASE remove useless comments when writing actual code. If needed then use Doxygen comments and tags.
     */

    /* Process all parameters from both command-line and .ini file */
    /* get the module name which will form the stem of all module port names */
    moduleName            = rf.check("name", 
                           Value("RopsciSequencer"), 
                           "module name (string)").asString();

    /*
    * before continuing, set the module name before getting any other parameters, 
    * specifically the port names which are dependent on the module name
    */
    setName(moduleName.c_str());

    /* get the name of the input and output ports, automatically prefixing the module name by using getName() */
    statusPortName        = "/";
    statusPortName       += getName(
                           rf.check("statusPort", 
                           Value("/RopsciSequencer/status:o"),
                           "Output port (string)").asString()
                           );

    jasPortName        = "/";
    jasPortName       += getName(
                           rf.check("jasPort", 
                           Value("/RopsciSequencer/jas:o"),
                           "Output port (string)").asString()
                           );

	esPortName        = "/";
    esPortName       += getName(
                           rf.check("esPort", 
                           Value("/RopsciSequencer/es:o"),
                           "Output port (string)").asString()
                           );

    filename        = rf.check("sequencefile",
                           Value("conf/RPSSequences.txt"),
                           "Key value (string)").asString();

	state = -1;
	step = 0;
	loop = false;
	executed = false;

	if (!this->loadSequences()) {
		cout << getName() << ": Unable to load sequences from " << filename << endl;  
        return false;
	}

    /* do all initialization here */
    /*
    * attach a port of the same name as the module (prefixed with a /) to the module
    * so that messages received from the port are redirected to the respond method
    */
    handlerPortName =  "/";
    handlerPortName += getName();         // use getName() rather than a literal 
	handlerPortName += "/command";
    if (!handlerPort.open(handlerPortName.c_str())) {           
        cout << getName() << ": Unable to open port " << handlerPortName << endl;  
        return false;
    }
    attach(handlerPort); // attach to port
    attachTerminal(); // attach to terminal

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;  
		return false;
	}

    /* create the thread and pass pointers to the module parameters */
    dThread = new RopsciSequencerThread(this);

    /* now start the thread to do the work */    
    dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

    return true ;     // let the RFModule know everything went well
                      // so that it will then run the module
}

bool RopsciSequencerModule::interruptModule() {
    handlerPort.interrupt();
    return true;
}

bool RopsciSequencerModule::close() {
    //close ports
    handlerPort.close();
    //portOut.interrupt();
	statusOut.close();
    jasOut.close();
	esOut.close();

    /* stop the thread */
    dThread->stop();

    return true;
}

bool RopsciSequencerModule::respond(const Bottle& command, Bottle& reply) {
  string helpMessage =  string(getName().c_str()) + 
                        " commands are: \n" +  
                        "help \n" + 
                        "quit \n" + 
                        "set thr <n> ... set the threshold \n" + 
                        "(where <n> is an integer number) \n";

  reply.clear(); 

  if (command.get(0).asString()=="quit") {
       reply.addString("quitting");
       return false;     
   }
   else if (command.get(0).asString()=="help") {
      cout << helpMessage;
      reply.addString("ok");
   }
   else if (command.get(0).asString()=="set") {
      if (command.get(1).asString() == "state") {
		int desiredstate = command.get(2).asInt();
		if (desiredstate >= (int) this->jointangleseqs.size()) {
			reply.addString("state exceeds number of loaded sequences");
		} else {
			//if (desiredstate == -1)	this->executed = true;
			this->state = desiredstate;
			this->step = 0;
			this->lastTimeSwitched = Time::now();//in seconds
			reply.addString("ok");
		}
	} else if (command.get(1).asString() == "step") {
		int desiredstep = command.get(2).asInt();
		if (desiredstep >= (int) this->jointangleseqs[this->state].size()) {
			reply.addString("step exceeds sequence length");
		} else {
			//TODO lock/pause thread
			this->step = desiredstep;
			this->executed = false;
			reply.addString("ok");
		}
	}

		//	
   }
   return true;
}

bool RopsciSequencerModule::openPorts() {
    if (!statusOut.open(statusPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }
	if (!jasOut.open(jasPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }
	if (!esOut.open(esPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }
	return true;
}

/* Called periodically every getPeriod() seconds */
bool RopsciSequencerModule::updateModule() {

    return true;
}

double RopsciSequencerModule::getPeriod() {
    /* module periodicity (seconds), called implicitly by myModule */    
    return 0.1;
}


bool RopsciSequencerModule::loadSequences() {
	// try to open file
	std::ifstream file(filename.c_str(), std::ifstream::in); //std::ios::in|std::ios::binary);
	if (!file.is_open()) {
		return false;
	}

	std::cout << "[RopsciSequencerModule::loadSequences] loading sequences from " << filename << " ..." << std::endl;
	while (file.good()) {
		//read sequence info/header
		std::string str;
		file >> str;
		if (str == "")	continue;
		seqnames.push_back(str);
		unsigned int steps;
		file >> steps;

		//read the joint angle vectors
		unsigned int dim;
		file >> dim;
		double tmp;
		std::vector<yarp::sig::Vector> tmpja;
		for (int i=0; i<steps; i++) {
			yarp::sig::Vector v(dim);
			file >> str;
			if (str == "none") {
				tmpja.push_back(yarp::sig::Vector(0));
			} else {
				v[0] = DEG2RAD * atof(str.c_str());
				for (int j=1; j<dim; j++) {
					file >> tmp;
					v[j] = DEG2RAD * tmp;
					//v[j] = tmp;
				}
				tmpja.push_back(v);
			}
		}
		jointangleseqs.push_back(tmpja);

		//read the step durations (in seconds)
		std::vector<double> tmps;
		for (int i=0; i<steps; i++) {
			file >> tmp;
			tmps.push_back(tmp);
		}
		pausesecs.push_back(tmps);
	
		//read the text pieces ("none" indicates that no text is spoken in that step)
		std::vector<std::string> tmpaf;
		for (int i=0; i<steps; i++) {
			//TODO getline !!!!
			file >> str;
			tmpaf.push_back(str);
		}
		voiceseqs.push_back(tmpaf);
	}

	std::cout << "[RopsciSequencerModule::loadSequences] loaded " << jointangleseqs.size() << " sequence(s)" << std::endl;

    return true;
}



