/**
 * @file SimpleGazeCtrl.cpp
 * @brief implementation of the SimpleGazeCtrlModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */ 

#include "iCub/SimpleGazeCtrlModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace yarp::math;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool SimpleGazeCtrlModule::configure(yarp::os::ResourceFinder &rf) {    
    /* Process all parameters from both command-line and .ini file */
    moduleName            = rf.check("name", 
                           Value("SimpleGazeCtrl"), 
                           "module name (string)").asString();
    setName(moduleName.c_str());

    robotName             = rf.check("robot", 
                           Value("icub"), 
                           "Robot name (string)").asString();

    bodypart         	= "head";

    inpMarkerPortName         = "/";
    inpMarkerPortName        += getName(
                           rf.check("inpMarkerPortName", 
                           Value("/SimpleGazeCtrl/marker:i"),
                           "Marker input port (string)").asString()
                           );

    outStatusPortName        = "/";
    outStatusPortName       += getName(
                           rf.check("outStatusPortName", 
                           Value("/SimpleGazeCtrl/status:o"),
                           "Status output port (string)").asString()
                           );

    handlerPortName =  "/";
    handlerPortName += getName();         // use getName() rather than a literal 
	handlerPortName += "/command";
    if (!handlerPort.open(handlerPortName.c_str())) {           
        cout << getName() << ": Unable to open port " << handlerPortName << endl;  
        return false;
    }
    attach(handlerPort); // attach to port
    attachTerminal(); // attach to terminal

    /* get the parameters */
	markerToTrack        = rf.check("markerToTrack",
                         	Value(0),
                         	"Key value (int)").asInt();
    k_pan        = rf.check("k_pan",
                         	Value(0.01),
                         	"Key value (float)").asDouble();
	k_tilt        = rf.check("k_tilt",
                         	Value(0.01),
                         	"Key value (float)").asDouble();
	k_rollcomp        = rf.check("k_rollcomp",
                         	Value(0.25),
                         	"Key value (float)").asDouble();
	d_home        = rf.check("d_home",
                         	Value(0.25),
                         	"Key value (float)").asDouble();
	d_lowpass 		= rf.check("d_lowpass",
                         	Value(0.9),
                         	"Key value (double)").asDouble();

	numStepsNoInp = 0;
	maxNumStepsNoInp = 10;

	inpsrc = SimpleGazeCtrlModule::MARKER;
	state = SimpleGazeCtrlModule::IDLE;

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;  
		return false;
	}

	//connect to robot
	success = openRobotInterface();
	if (!success) {
		cout << getName() << ": Unable to connect to robot " << robotName << endl;  
		return false;
	}

	//init variables
	int joints;
	iencs->getAxes(&joints);
	qhead.resize(joints);
	dqhead.resize(joints);
	dqheadOld.resize(joints);

	qhead.zero();
	dqhead.zero();
	dqheadOld.zero();

    /* create the thread and pass pointers to the module parameters */
    dThread = new SimpleGazeCtrlThread(this);

    /* now start the thread to do the work */    
    dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

    return true ;     // let the RFModule know everything went well
                      // so that it will then run the module
}

bool SimpleGazeCtrlModule::interruptModule() {
	qhead.zero();
	dqhead.zero();
	dqheadOld.zero();

    handlerPort.interrupt();
	inpMarker.interrupt();
	outStatus.interrupt();
    return true;
}

bool SimpleGazeCtrlModule::close() {
	//close robot
	ivel->stop();
	polydriver.close();

    //close ports
    handlerPort.close();

    inpMarker.interrupt();
	inpMarker.close();
    outStatus.interrupt();
	outStatus.close();

    /* stop the thread */
    dThread->stop();

    return true;
}

bool SimpleGazeCtrlModule::respond(const Bottle& command, Bottle& reply) {
	string helpMessage = string(getName().c_str()) + 
						" commands are: \n" +  
						"help \n" + 
						"quit \n" + 
						"set <param> <value> ... where <param> is the name of a parameter and <value> the value it should take";
						"get <param> ... where <param> the name of a parameter";

	reply.clear(); 
	if (command.get(0).asString()=="quit") {
		reply.addString("quitting");
		return false;
	} else if (command.get(0).asString() == "help") {
		cout << helpMessage;
		reply.addString("ok");
	} else if (command.get(0).asString() == "set") {
		if (command.get(1).asString() == "k_pan") {
			k_pan = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "k_tilt") {
			k_tilt = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "k_rollcomp") {
			k_rollcomp = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "d_home") {
			d_home = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "d_lowpass") {
			d_lowpass = command.get(2).asDouble();
			reply.addString("ok");
		} else if (command.get(1).asString() == "state") {
			state = command.get(2).asInt();
			if (command.get(2).asInt() == SimpleGazeCtrlModule::IDLE) {
				state = SimpleGazeCtrlModule::IDLE;
				dqheadOld.zero();
				bool ok = ivel->velocityMove(dqheadOld.data());
				ivel->stop();
				sleep(0.5);
				ivel->stop();
			}
			reply.addString("ok");
		} else if (command.get(1).asString() == "markerToTrack") {
			markerToTrack = command.get(2).asInt();
			reply.addString("ok");
		} else if (command.get(1).asString() == "inpsrc") {
			inpsrc = command.get(2).asInt();
			reply.addString("ok");
		} else {
			reply.addString("parameter not known");
		}
	} else if (command.get(0).asString() == "get") {
		if (command.get(1).asString()=="k_pan") {
			reply.addDouble(k_pan);
		} else if (command.get(1).asString()=="k_tilt") {
			reply.addDouble(k_tilt);
		} else if (command.get(1).asString()=="k_rollcomp") {
			reply.addDouble(k_rollcomp);
		} else if (command.get(1).asString()=="d_home") {
			reply.addDouble(d_home);
		} else if (command.get(1).asString()=="d_lowpass") {
			reply.addDouble(d_lowpass);
		} else {
			reply.addString("parameter not known");
		}
	}
	return true;
}

bool SimpleGazeCtrlModule::openPorts() {
    if (!inpMarker.open(inpMarkerPortName.c_str())) {
        cout  << ": unable to open output port " << inpMarkerPortName << endl;
        return false;
    }
    if (!outStatus.open(outStatusPortName.c_str())) {
        cout  << ": unable to open output port " << outStatusPortName << endl;
        return false;
    }
	return true;
}

bool SimpleGazeCtrlModule::openRobotInterface() {
	Property options;
	options.put("device", "remote_controlboard");
	std::string localportname        = "/";
    localportname       += getName();
	localportname       += "/" + robotName + "/";
	localportname       += bodypart;
	options.put("local", localportname.c_str());
	std::string robotportname = "/";
	robotportname += robotName;
	robotportname += "/";
	robotportname += bodypart;
	options.put("remote", robotportname.c_str());

	polydriver.open(options);
	polydriver.view(iencs);
	polydriver.view(ivel);

	if ( (!iencs) || (!ivel) ) {
		return false;
	} else {
		return true;
	}
}

/* Called periodically every getPeriod() seconds */
bool SimpleGazeCtrlModule::updateModule() {
	bool gohome = false;
	if (state == SimpleGazeCtrlModule::IDLE)
		return true;

	//read marker input
	Bottle *b = 0;
	int cnt = 0;
	b = inpMarker.read(true);
	/*while ((b == NULL) && (cnt < 1000000)) {
		b = inpMarker.read(false);
		sleep(0.1);
		cnt++;
	}
	if (cnt < 1000000)	std::cout << cnt << endl;*/
	if (b != NULL) {
		int found = -1;
		double pxlX = 0;
		double pxlY = 0;
		if (inpsrc == SimpleGazeCtrlModule::MARKER) {
			//check whether desired marker is tracked
			for (int i=1; i<b->size(); i+=3) {
				if (b->get(i).asInt() == markerToTrack) {
					found = i;
					//break;
				}
			}
			if (found >= 0) {
				//printf("found markerToTrack\n");
				pxlX = b->get(found+1).asDouble();
				pxlY = b->get(found+2).asDouble();
			}
		} else {
			//"total -8.0 30.0 1.0"
			pxlX = b->get(1).asDouble();
			pxlY = b->get(2).asDouble();
			double confidence = b->get(3).asDouble();
			if (confidence > 0.75) {
				//printf("something to track\n");
				found = 1;
			}
		}

		//gaze at the marker/pixel
		if (found >= 0) {
			numStepsNoInp = 0;

			double dpan = -k_pan * pxlX;
           	double dtilt = -k_tilt * pxlY;
			//cout << pxlX << " " << k_pan << std::endl;

			//Roll Compensation
			iencs->getEncoders(qhead.data());
			double tilt = DEG2RAD * qhead[0];
			double roll = DEG2RAD * qhead[1];
			double pan = DEG2RAD * qhead[2];
	        double desiredRoll = rollCompensation(pan, tilt);
			//double desiredRoll = 0;
			double droll = k_rollcomp * (desiredRoll - roll);

			dqhead.zero();
			dqhead[0] = RAD2DEG * dtilt;
			dqhead[1] = RAD2DEG * droll;
			dqhead[2] = RAD2DEG * dpan;
			dqheadOld = (1.0 - d_lowpass) * dqheadOld + d_lowpass * dqhead;
			bool ok = ivel->velocityMove(dqheadOld.data());
			//std::cout << ok << " " << dqhead.toString() << std::endl;
		} else {
			gohome = true;
			numStepsNoInp++;
		}
	} else {
		gohome = true;
		numStepsNoInp++;
	}

	//accept no input for a few steps
	//std::cout << numStepsNoInp << std::endl;
	if (numStepsNoInp > maxNumStepsNoInp) {
		gohome = true;
		//std::cout << "lost target" << std::endl;
	} else {
		gohome = false;
	}

	//simple control to home posture (0, 0)
	if (gohome) {
		dqhead.zero();
		iencs->getEncoders(qhead.data());
		for (int i=dqhead.size()-1; i>=0; i--) {
			dqhead[i] = -d_home * qhead[i];
		}
		dqheadOld = (1.0 - d_lowpass) * dqheadOld + d_lowpass * dqhead;
		bool ok = ivel->velocityMove(dqheadOld.data());
		//std::cout << "going home" << endl; //<< ok << " " << dqhead.toString() << std::endl;

		//ivel->stop();
		//bool ok = ivel->velocityMove(dqhead.data());
	}

    return true;
}

double SimpleGazeCtrlModule::rollCompensation(double pan, double tilt) {
    return asin( (sin(pan)*sin(tilt))/(cos(pan)*cos(tilt)) );
}

double SimpleGazeCtrlModule::getPeriod() {
    /* module periodicity (seconds), called implicitly by myModule */    
    return 0.1;
}




