/**
 * @file iCubTemplateModule.cpp
 * @brief implementation of the iCubTemplateModuleModule methods following the RFModule standards.
 */

/*
 * Audit Trail
 * -----------
 *  .... if any
 */ 

#include "iCub/iCubTemplateModuleModule.h"

using namespace yarp::os;
using namespace yarp::sig;
using namespace std;

/* 
 * Configure method. Receive a previously initialized
 * resource finder object. Use it to configure your module.
 * If you are migrating from the old Module, this is the 
 * equivalent of the "open" method.
 */

bool iCubTemplateModuleModule::configure(yarp::os::ResourceFinder &rf) {    
    /*
     * PLEASE remove useless comments when writing actual code. If needed then use Doxygen comments and tags.
     */

    /* Process all parameters from both command-line and .ini file */
    /* get the module name which will form the stem of all module port names */
    moduleName            = rf.check("name", 
                           Value("iCubTemplateModule"), 
                           "module name (string)").asString();

    /*
    * before continuing, set the module name before getting any other parameters, 
    * specifically the port names which are dependent on the module name
    */
    setName(moduleName.c_str());

    /* now, get the rest of the parameters */
    /*
    * get the robot name which will form the stem of the robot ports names
    * and append the specific part and device required
    */
    robotName             = rf.check("robot", 
                           Value("icub"), 
                           "Robot name (string)").asString();

    robotPortName         = "/" + robotName + "/head";

   

    /* get the name of the input and output ports, automatically prefixing the module name by using getName() */
    inputPortName         = "/";
    inputPortName        += getName(
                           rf.check("myInputPort", 
                           Value("/image:i"),
                           "Input image port (string)").asString()
                           );

    outputPortName        = "/";
    outputPortName       += getName(
                           rf.check("myOutputPort", 
                           Value("/image:o"),
                           "Output image port (string)").asString()
                           );

    /* get the threshold value */
    thresholdValue        = rf.check("threshold",
                           Value(8),
                           "Key value (int)").asInt();


    /* do all initialization here */
    /*
    * attach a port of the same name as the module (prefixed with a /) to the module
    * so that messages received from the port are redirected to the respond method
    */
    handlerPortName =  "/";
    handlerPortName += getName();         // use getName() rather than a literal
	handlerPortName += "/command"; 
    if (!handlerPort.open(handlerPortName.c_str())) {           
        cout << getName() << ": Unable to open port " << handlerPortName << endl;  
        return false;
    }
    attach(handlerPort); // attach to port
    attachTerminal(); // attach to terminal

	//open ports
	bool success = openPorts();
	if (!success) {
		cout << getName() << ": Unable to open ports" << endl;  
		return false;
	}

    /* create the thread and pass pointers to the module parameters */
    dThread = new iCubTemplateModuleThread(this);

    /* now start the thread to do the work */    
    dThread->start(); // this calls threadInit() and it if returns true, it then calls run()

    return true ;     // let the RFModule know everything went well
                      // so that it will then run the module
}

bool iCubTemplateModuleModule::interruptModule() {
    handlerPort.interrupt();
    return true;
}

bool iCubTemplateModuleModule::close() {
    //close ports
    handlerPort.close();

    //imagePortIn.interrupt();
    //portOut.interrupt();
    //imagePortIn.close();
    //portOut.close();

    /* stop the thread */
    dThread->stop();

    return true;
}

bool iCubTemplateModuleModule::respond(const Bottle& command, Bottle& reply) {
  string helpMessage =  string(getName().c_str()) + 
                        " commands are: \n" +  
                        "help \n" + 
                        "quit \n" + 
                        "set thr <n> ... set the threshold \n" + 
                        "(where <n> is an integer number) \n";

  reply.clear(); 

  if (command.get(0).asString()=="quit") {
       reply.addString("quitting");
       return false;     
   }
   else if (command.get(0).asString()=="help") {
      cout << helpMessage;
      reply.addString("ok");
   }
   else if (command.get(0).asString()=="set") {
      if (command.get(1).asString()=="thr") {
         thresholdValue = command.get(2).asInt(); // set parameter value
         reply.addString("ok");
      }
   }
   return true;
}

bool iCubTemplateModuleModule::openPorts() {
    /*if (!imagePortIn.open(inputPortName.c_str()) ) {
        std::cout << ": unable to open input port " << std::endl;
        return false;  // unable to open; let RFModule know so that it won't rrun
    }*/

    if (!portOut.open(outputPortName.c_str())) {
        cout  << ": unable to open output port " <<  endl;
        return false;  // unable to open; let RFModule know so that it won't run
    }

	return true;
}

/* Called periodically every getPeriod() seconds */
bool iCubTemplateModuleModule::updateModule() {

    return true;
}

double iCubTemplateModuleModule::getPeriod() {
    /* module periodicity (seconds), called implicitly by myModule */    
    return 0.1;
}




