classdef rpsNet < handle
    
    properties
        
        par;    % network parameters
        log;    % logging variables
        graph;  % graphic objects
        
        %% simulation state variables
        neuralOutputs;                  % nrNeurons x 1 vector holding current step neural output
        neuralOutputsPrevious;          % similar to neuralOutput but remembers the values of the
        % previous step. Necessary to compute the Hebbian product
        % 1-step episodes with such minimum interval
        stimuliInputToNeurons;          % nrNeurons x 1 vector feeding neurons with external input
        stimuliNeuronsGroups;           % nrStimuli x nrNeuronsInGroup vector
        stimuliProcessed;               % nrStimuli x 1 binary vector of active inputs
        % at the corrent time step (after processing)
        stimuliRaw;                     % nrStimuli x 1 binary vector of active inputs
        % at the corrent time step (before processing)
        rewardRaw;                      % binary value of presence of reward from the env.
        rewardRefractoryTime;           % minimum interval between two reward episodes
        outputNeuronsGroups;            % neurons belonging to the output groups
        actions;
        CS;                             % 1 when a CS is triggered
        stimulusToTrack;
        neuronInputs;
        output = 0;
        
        connectivity;                   % nrNeurons^2 sparse connectivity matrix
        neuronTypes;                    % nrNeurons vector of neuron types (excit/inhib)
        connectionType;
        connectionTypeStrength;
        weightMatrix;
        shortTermComponent;
        longTermComponent;
        correlations;
        highCorr;
        plasticityMatrix;
        nrOfSynapses;
        nrOfPlasticSynapses;
        
        modulation;
        stream;
        
        FIFOmonitoringWindowOfCorrelations;
        FIFOmonitoringWindowOfDecorrelations;
        corSamples;
        thresholdCorSamples;
        thresholdDecSamples;
        nrOfCorPerStep;
        nrOfDecPerStep;
        RCpercentPerSecond;
        RDpercentPerSecond;
        thresholdCor;
        thresholdDec;
        updateThreshold;
        nrStepsInMonitoringWindow;
        colorStimulusPresentFor = 0;
        thinkingColor = 0;
        winner = 0;
        outvalues;
        totalSteps;
        currentStep;
        cputime;
        
        
    end
    
    methods
        function net = rpsNet(seed, nrStimuli, nrOutputs, sampling, maxTime, conditioningType)
            if sampling > 1
                error('The sampling time must be less than or equal to 1 s');
            end
            if rem(1,sampling) > 0
                error('The sampling step must be set such that 1 is a multiple');
            end
            if nrStimuli > 12
                error('For non-overlapping groups, 10 groups is the maximum suggested nr of stimuli');
            end
            
            disp('Initialising...');
            net.par.seedNumber = seed;
            
            %%#############################################################
            % Parameter settings 
            %##############################################################
  
            net.par.samplingTime                        = sampling;
            net.par.conditioningType                    = conditioningType;
            net.par.shortTC                             =  2.0;            % traces time constant
            net.par.modTC                               =  1.0;            % mod time constant
            net.par.inhibWeight                         =  0.5;            % fixed weight of inhibitory synapses
            
            net.par.targetRCPercentPerSecond            =  0.5;            % rare correlation target rate
            net.par.corrTollerance                      =  5;              % multiplication/division factor for upper and lower tolerance 
            net.par.durationMonitoringWindow            =  10.0;
            net.par.thetaUpdateRate                     =  0.002;          % per sec
            
            net.par.maxWeight                           =  1.0;
            net.par.minWeight                           =  0.0;
            net.par.weightInitRange                     =  0.1;
            net.par.transmissionNoise                   =  0.05;            % Xi in Eq. 6
            net.par.regularConnectivity                 =  0;              % value 1 initialises the networks with regular connectivity
            net.par.traceIncrementWithCor               =  0.1;            % alfa in Eq. 1
            net.par.traceDecrementWithDec               = -0.1;            % beta in Eq. 1
            
            net.par.nrNeurons                           = 800;
            net.par.nrExcitatory                        = 620;
            net.par.nrInhibitory                        = net.par.nrNeurons - net.par.nrExcitatory;
            net.par.probConnect                         = 0.1;
            
            net.par.k                                   = [1 -5];          % Kappa in Eq. 5
            
            net.par.activationFactor                    = 0.2;            % Gamma in Eq. 6
            
            net.par.maxModulation                       = 0.7;            % during reward episodes
            net.par.baseModulation                      = -0.00;          % always present
            net.par.outputContributionToMod             = -0.00;           % output increases the level of modulation
            
            net.par.stimuliStrength                     = 10;
            net.par.strengthOfFeedback                  = 0.05;
            
            net.par.nrNeuronsInGroup                    = 50;              % number of neurons in an input group
            net.par.nrNeuronsOutGroup                   = 50;              % number of neurons in an output group
            net.par.conditionedResponseTriggeringThreshold = 0.5;
            
            net.par.colourImmediatelyRecognisedThreshold   = 0.2;
            net.par.observeColorFor                     = 0;  
            net.par.thinkColorForMax                    = 1.5;
            
            %% initialisation of state variables
            
            net.stream = RandStream('mrg32k3a', 'Seed', net.par.seedNumber);
                        
            net.par.nrStimuli                           = nrStimuli;
            net.par.nrOutputs                           = nrOutputs;
            net.initialiseNeuronGroups();

            
            net.stimuliRaw              = zeros(net.par.nrStimuli,1);
            net.stimuliProcessed        = zeros(net.par.nrStimuli,1);
            net.stimuliInputToNeurons   = zeros(net.par.nrNeurons,1);
            net.par.stepsPerSecond                      = 1/net.par.samplingTime;
            
            net.neuralOutputs           = net.mysymrand(net.par.nrNeurons, 1, net.par.transmissionNoise);
            net.neuralOutputsPrevious   = zeros(net.par.nrNeurons, 1);
            
            net.initialiseConnectivity();
            
            net.neuronTypes = [ones(1,net.par.nrExcitatory) -ones(1,net.par.nrInhibitory)];
            
            net.connectionType = net.connectivity .* repmat(net.neuronTypes, net.par.nrNeurons, 1)';
            
            net.connectionTypeStrength  = sparse(net.par.nrNeurons, net.par.nrNeurons);
            net.connectionTypeStrength  = net.connectionTypeStrength + net.connectivity .* net.par.k(2) .* (net.connectionType == -1);
            net.connectionTypeStrength  = net.connectionTypeStrength + net.connectivity .* net.par.k(1) .* (net.connectionType == 1);
            
            net.neuronInputs            = zeros(net.par.nrNeurons,1);
            
            net.weightMatrix            = sparse(net.par.nrNeurons,net.par.nrNeurons);
            net.longTermComponent       = sparse(net.par.nrNeurons,net.par.nrNeurons);
            net.shortTermComponent      = sparse(net.par.nrNeurons,net.par.nrNeurons);
            
            net.longTermComponent       = rand(net.stream, net.par.nrNeurons, net.par.nrNeurons)...
                .* net.par.maxWeight .* net.par.weightInitRange .* net.connectivity;
            
            net.longTermComponent(net.par.nrExcitatory + 1:net.par.nrNeurons,:) = ...
                net.par.maxWeight .* net.par.inhibWeight .* net.connectivity(net.par.nrExcitatory + 1:net.par.nrNeurons,:);
            
            net.weightMatrix            = net.longTermComponent + net.shortTermComponent;
            
            net.plasticityMatrix = net.connectivity;
            net.plasticityMatrix(net.par.nrExcitatory + 1:net.par.nrNeurons,:) = 0;
            
            % new code 21/01
            %net.plasticityMatrix(net.outputNeuronsGroups(:),:) = 0;
            %net.plasticityMatrix(:,net.stimuliNeuronsGroups(:)) = 0;
            %
            
            net.nrOfSynapses            = nnz(net.connectivity);
            net.nrOfPlasticSynapses     = nnz(net.plasticityMatrix);
            
            net.nrStepsInMonitoringWindow           = net.par.durationMonitoringWindow /net.par.samplingTime;
            
            net.thresholdCor            = 0.1;
            net.thresholdDec            = -0.1;
                        
            net.FIFOmonitoringWindowOfCorrelations         = zeros(net.nrStepsInMonitoringWindow,1);
            net.FIFOmonitoringWindowOfDecorrelations         = zeros(net.nrStepsInMonitoringWindow,1);
                        
            net.totalSteps = 1/net.par.samplingTime * maxTime;
            net.currentStep             = 3;
            net.modulation              = 0;
            net.stimulusToTrack         = 0;
            net.rewardRefractoryTime    = 0;
            net.CS                      = 0;
            net.cputime                 = zeros(9,1);
            net.outvalues = zeros(net.par.nrOutputs,1);

            
            net.log.modulation          = zeros(1, net.totalSteps);
            net.log.RCrate              = zeros(1, net.totalSteps);
            net.log.RDrate              = zeros(1, net.totalSteps);
            net.log.thresholdCor        = zeros(1, net.totalSteps);
            net.log.thresholdDec        = zeros(1, net.totalSteps);
            net.log.allSynapses         = zeros(1, net.totalSteps);
            net.log.synapsesS           = zeros(net.totalSteps,net.par.nrStimuli);
            
            for i = 1:net.par.nrStimuli
                net.log.synapsesStoO{i}         = zeros(net.totalSteps,net.par.nrOutputs);
            end
            
            
            net.log.meanEW              = zeros(1, net.totalSteps);
            net.log.meanIW              = zeros(1, net.totalSteps);
            net.log.activityOG          = zeros(net.totalSteps, net.par.nrOutputs);
            net.log.activityIG          = zeros(net.totalSteps, net.par.nrStimuli);
            net.log.stimuliRaw          = zeros(net.totalSteps, net.par.nrStimuli);
            net.log.stimuliProcessed    = zeros(net.totalSteps, net.par.nrStimuli);
            net.log.stimulus1           = zeros(1, net.totalSteps);
            net.log.actions             = zeros(net.totalSteps, net.par.nrOutputs);
            net.log.smile               = zeros(net.totalSteps, 1);
            net.log.stimulusToTrack     = zeros(net.totalSteps, 1);
            net.log.nrCurrentStimuli    = zeros(net.totalSteps, 1);
            net.log.totC                = zeros(net.totalSteps, 1);
            net.log.C_S1toO             = zeros(net.totalSteps, 1);
            disp('Initialisation complete.');
        end
        
        
        function [] = update(net, stimuli, reward, verbose)
            if nargin < 4
                verbose = 0;
            end
            
            net.stimuliRaw = stimuli;
            net.rewardRaw = reward;
            
            tic;
            net.processStimuliCont();
            net.cputime(1) = toc;
            
            tic
            net.neuralComputation;
            net.cputime(2) = toc;
            
            tic
            if net.par.conditioningType == 1
                net.readCS;
            else
                net.readAction;
            end
            net.cputime(3) = toc;
                        
            tic
            net.monitorCorrelations;
            net.cputime(5) = toc;
            
            tic
            net.plasticity;
            net.cputime(7) = toc;
            
            tic
            net.updateTraces;
            net.cputime(8) = toc;
            
            tic
            net.logCurrentStep;
            net.cputime(9) = toc;
            
            net.currentStep = net.currentStep + 1;
            if verbose
                disp(['step time: ' num2str(sum(net.cputime)) ' : ' num2str(net.cputime')]);
            end
        end
        
       
      function [] = processStimuliCont(net)
          net.stimuliProcessed = net.stimuliRaw;
          
          net.output = sum(sum(net.neuralOutputs(net.outputNeuronsGroups(:,:))))...
              /size(net.outputNeuronsGroups,2);
          
          net.stimuliInputToNeurons(net.stimuliNeuronsGroups(1:net.par.nrStimuli,:)) = - 0.5 * net.par.stimuliStrength .* net.par.samplingTime;
          
          net.stimuliInputToNeurons(net.stimuliNeuronsGroups(net.stimuliProcessed >= 1,:)) =  net.par.stimuliStrength .* net.par.samplingTime;
          
          net.modulation =  net.modulation .* exp(-net.par.samplingTime/net.par.modTC)...
              + net.par.baseModulation * net.par.samplingTime...
              + net.par.maxModulation * net.rewardRaw; %...
          
          net.modulation = max(net.modulation,-3 *net.par.maxModulation);
            %  + net.par.outputContributionToMod * net.output...
            %  - 0.001 * rand() * net.par.samplingTime;
                    
      end
      
      
      function [] = neuralComputation(net)
          
          net.neuralOutputsPrevious = net.neuralOutputs;
          
          % this line for eLTP as eligibility traces
          net.neuronInputs = net.stimuliInputToNeurons' + net.neuralOutputs' * (net.weightMatrix .* net.connectionTypeStrength);
          % this line for separate values for eligibility traces, i.e. not used for
          % computing neural trasmission
          %net.neuronInputs = net.stimuliInputToNeurons' + net.neuralOutputs' * (net.longTermComponent .* net.connectionTypeStrength);
          
          % computing new outputs and adding noise
          net.neuralOutputs =  max(tanh(net.par.activationFactor .* net.neuronInputs),0.0)'...
              + net.mysymrand(net.par.nrNeurons, 1, net.par.transmissionNoise);
          
          net.correlations = sparse(net.plasticityMatrix .* (net.neuralOutputsPrevious * net.neuralOutputs'));
          
          % resetting high correlationa from previous time step
          net.highCorr = sparse(net.par.nrNeurons,net.par.nrNeurons);
          
          % finding correlating weights
          net.highCorr(net.correlations > net.thresholdCor) = net.par.traceIncrementWithCor;
          net.nrOfCorPerStep = nnz(net.highCorr);
          
          % finding decorrelating weights
          net.highCorr(net.correlations < net.thresholdDec) = net.par.traceDecrementWithDec;
          net.nrOfDecPerStep = nnz(net.highCorr == net.par.traceDecrementWithDec);
          
      end
      
      function [] = monitorCorrelations(net)
          % Shifting FIFO
          % case in which the monitorying window has more than one sample, i.e. nearly
          % always
          if net.nrStepsInMonitoringWindow > 1
              net.FIFOmonitoringWindowOfCorrelations(2:net.nrStepsInMonitoringWindow) =...
                  net.FIFOmonitoringWindowOfCorrelations(1:net.nrStepsInMonitoringWindow - 1);
              net.FIFOmonitoringWindowOfDecorrelations(2:net.nrStepsInMonitoringWindow) =...
                  net.FIFOmonitoringWindowOfDecorrelations(1:net.nrStepsInMonitoringWindow - 1);
              
          end
          net.FIFOmonitoringWindowOfCorrelations(1) = net.nrOfCorPerStep;
          net.FIFOmonitoringWindowOfDecorrelations(1) = net.nrOfDecPerStep;
          
          net.RCpercentPerSecond = sum(net.FIFOmonitoringWindowOfCorrelations) / (net.nrOfPlasticSynapses * net.par.durationMonitoringWindow) * 100;
          net.RDpercentPerSecond = sum(net.FIFOmonitoringWindowOfDecorrelations) / (net.nrOfPlasticSynapses * net.par.durationMonitoringWindow) * 100;
          
          % too many correlations
          if net.RCpercentPerSecond > net.par.corrTollerance * net.par.targetRCPercentPerSecond
              net.thresholdCor = net.thresholdCor + net.par.thetaUpdateRate * net.par.samplingTime ;
          end
          % too few correlations
          if  net.RCpercentPerSecond < 1/net.par.corrTollerance * net.par.targetRCPercentPerSecond
              net.thresholdCor = net.thresholdCor - net.par.thetaUpdateRate * net.par.samplingTime;
          end
          % too many dec
          if net.RDpercentPerSecond > net.par.corrTollerance * net.par.targetRCPercentPerSecond
              net.thresholdDec = net.thresholdDec - net.par.thetaUpdateRate * net.par.samplingTime;
          end
          % too few dec
          if  net.RDpercentPerSecond < 1/net.par.corrTollerance * net.par.targetRCPercentPerSecond
              net.thresholdDec = net.thresholdDec + net.par.thetaUpdateRate * net.par.samplingTime;
          end
      end
      
      
      function [] = plasticity(net)
          
          net.longTermComponent(1:net.par.nrExcitatory,:) = sparse(net.longTermComponent(1:net.par.nrExcitatory,:)...
              +  (net.shortTermComponent(1:net.par.nrExcitatory,:) .* net.modulation));
          
          % Weights clipping
          net.longTermComponent = sparse(max(min(net.longTermComponent, net.par.maxWeight), net.par.minWeight));
          net.weightMatrix = sparse(max(min((net.longTermComponent + net.shortTermComponent), net.par.maxWeight), net.par.minWeight));
      end
      
      function [] = updateTraces(net)
          
          net.shortTermComponent = sparse (  (net.shortTermComponent )...
              .* exp(-net.par.samplingTime/net.par.shortTC) + net.highCorr  );
          
          % clipping
          net.shortTermComponent = max(min(net.shortTermComponent, net.par.traceIncrementWithCor), net.par.traceDecrementWithDec);
      end
      
      function [] = logCurrentStep(net)
          net.log.modulation(net.currentStep) = net.modulation;
          net.log.RCrate(net.currentStep) = net.RCpercentPerSecond;
          net.log.RDrate(net.currentStep) = net.RDpercentPerSecond;
          net.log.thresholdCor(net.currentStep) = net.thresholdCor;
          net.log.thresholdDec(net.currentStep) = net.thresholdDec;
          
          net.log.allSynapses(net.currentStep) = sum(sum(net.longTermComponent(1:net.par.nrExcitatory,:)))...
              /nnz(net.connectivity(1:net.par.nrExcitatory,:));
          for i = 1:net.par.nrStimuli
              net.log.synapsesS(net.currentStep,i) = sum(sum(net.longTermComponent(net.stimuliNeuronsGroups(i,:),:)))...
                  /nnz(net.connectivity(net.stimuliNeuronsGroups(i,:),:));
              net.log.activityIG(net.currentStep,i) = sum(net.neuralOutputs(net.stimuliNeuronsGroups(i,:)))/size(net.stimuliNeuronsGroups,2);
              for j = 1:net.par.nrOutputs
                  net.log.synapsesStoO{i}(net.currentStep,j) =...
                      sum(sum(net.longTermComponent(net.stimuliNeuronsGroups(i,:),net.outputNeuronsGroups(j,:))))...
                      /nnz(net.connectivity(net.stimuliNeuronsGroups(i,:),net.outputNeuronsGroups(j,:)));
              end
     
          end
          for i = 1:net.par.nrOutputs
              net.log.activityOG(net.currentStep,i) = sum(net.neuralOutputs(net.outputNeuronsGroups(i,:)))/size(net.outputNeuronsGroups,2);
          end
          
          net.log.stimuliRaw(net.currentStep,:) = net.stimuliRaw .* (1:net.par.nrStimuli)';
          net.log.stimuliProcessed(net.currentStep,:) = min(net.stimuliProcessed,1) .* (1:net.par.nrStimuli)';
          net.log.nrCurrentStimuli(net.currentStep) = sum(net.stimuliRaw);
          
          if net.par.conditioningType == 2
              net.log.actions(net.currentStep,:) = net.actions .* (1:net.par.nrOutputs)';
                 
          end
          if net.par.conditioningType == 1
              net.log.smile(net.currentStep) = net.CS;
              net.log.stimulusToTrack(net.currentStep) = net.stimulusToTrack;
          end
          
          % Extra log (slow) only for fig. in the paper
          % amount of traces along S*->GA0
          net.log.totC(net.currentStep) = sum(sum(net.shortTermComponent));
          net.log.C_S1toO(net.currentStep) = sum(sum(net.shortTermComponent(net.stimuliNeuronsGroups(1,:),net.outputNeuronsGroups(1,:))));
          
          
      end
      
      function [] = readCS(net)
          net.CS = 0;
          net.stimulusToTrack = -1;
          
          if  net.output > net.par.conditionedResponseTriggeringThreshold
              net.CS = 1;
              % adjustment to cancel out negative effect of output on modulation when CS
              % is present
              net.modulation = net.modulation - net.par.outputContributionToMod * net.output;
          end
          
          if net.CS
              tmp = find(net.log.stimuliProcessed(net.currentStep - 2, :) > 0);
              if ~isempty(tmp)
                  net.stimulusToTrack = tmp(1);
              end
          end
      end
      
      
      function [] = readAction(net)
          net.actions = zeros(net.par.nrOutputs,1);
          % if one of the two groups receives an input
          inputValues = zeros(net.par.nrStimuli,1);
          for i = 1:net.par.nrStimuli
            %  inputValues(i) = sum(net.neuralOutputs(net.stimuliNeuronsGroups(i,:)))/size(net.stimuliNeuronsGroups,2) > 0.2...
            %      || net.log.activityIG(max(1,net.currentStep),2);
            % The above bug was found on 30/01: what was the effect? it
            % apperas that input values of i were always from input group 2
            % if input group i was less than 0.2
            % the threshold value was also not relating to the  correct
            % parameter above
              inputValues(i) = sum(net.neuralOutputs(net.stimuliNeuronsGroups(i,:)))/size(net.stimuliNeuronsGroups,2)...
                  > net.par.colourImmediatelyRecognisedThreshold ...
                  || net.log.activityIG(max(1,net.currentStep),i);
          end
          
          % measuring duration of stimuli
          if find(net.stimuliProcessed > 0) > 0
              net.colorStimulusPresentFor = net.colorStimulusPresentFor + 1;
          else
              net.colorStimulusPresentFor = 0;
              net.winner = 0;             
          end
          
          if net.thinkingColor == 0
              net.outvalues = zeros(net.par.nrOutputs,1);
          end
          
          if net.colorStimulusPresentFor * net.par.samplingTime > net.par.observeColorFor
              net.thinkingColor = net.thinkingColor + 1;
              for i = 1:net.par.nrOutputs
                  net.outvalues(i) = 0.75 * net.outvalues(i) + sum(net.neuralOutputs(net.outputNeuronsGroups(i,:)))...
                      /size(net.outputNeuronsGroups,2);
              end
              if net.winner == 0
                  if (max(net.outvalues) > net.par.colourImmediatelyRecognisedThreshold || net.thinkingColor > net.par.thinkColorForMax / net.par.samplingTime)
                      net.winner = find(net.outvalues == max(net.outvalues));
                      disp(['max out ' num2str(max(net.outvalues)) ' winner is ' num2str(net.winner)]);
                  else
                      %disp(['max out ' num2str(max(net.outvalues)) ' thinking for ' num2str(net.thinkingColor)]);
                  end
              end
              if net.winner ~= 0 % there is a winner
                  net.thinkingColor = 0;

                 
                  %otherActions = ones(net.par.nrOutputs,1);
                  %               otherActions(net.winner) = 0;
                  %net.stimuliInputToNeurons(net.outputNeuronsGroups(otherActions == 1,:),:) =  -1 * net.par.stimuliStrength;
                  net.stimuliInputToNeurons(net.par.nrStimuli*net.par.nrNeuronsInGroup:net.par.nrExcitatory ,:) =  -net.par.strengthOfFeedback * net.par.stimuliStrength;
                  
                  net.actions(net.winner) = 1;
                  net.stimuliInputToNeurons(net.outputNeuronsGroups(net.winner,:)) = net.par.strengthOfFeedback * net.par.stimuliStrength;
%  
              end
          else
              net.stimuliInputToNeurons(net.outputNeuronsGroups(:)) = 0;
          end
      end
      
      function createWindow(net)
          
          net.graph.f1 = figure('Position',[0 0 1000 600]);
          
          net.graph.sp1 = subplot(4,2,1,'Parent',net.graph.f1);
          ylabel(net.graph.sp1,'S afferents');
          
          net.graph.sp2 = subplot(4,2,2,'Parent',net.graph.f1);
          net.graph.sp2_plot =  plot(net.log.modulation,'Parent',net.graph.sp2);
          set(net.graph.sp2_plot,'YDataSource','app.netCC.log.modulation');
          set(net.graph.sp2,'Ylim',[-0.1, net.par.maxModulation * 1.2]);
          grid;
          ylabel(net.graph.sp2,'modulation');
          
          net.graph.sp3 = subplot(4,2,3,'Parent',net.graph.f1);
          net.graph.sp3_plot1 = plot(net.log.thresholdCor,'Parent', net.graph.sp3);
          hold on
          net.graph.sp3_plot2 = plot(net.log.thresholdDec,'Parent', net.graph.sp3);
          set(net.graph.sp3_plot1,'YDataSource','app.netCC.log.thresholdCor');
          set(net.graph.sp3_plot2,'YDataSource','app.netCC.log.thresholdDec');
          ylabel(net.graph.sp3,'thresholds');
          set(gca,'Xlim',[1 net.currentStep]);
          grid;
          
          net.graph.sp4 = subplot(4,2,4,'Parent',net.graph.f1);
          
          net.graph.sp5 = subplot(4,2,5,'Parent',net.graph.f1);
          net.graph.sp5_plot = plot(net.log.RCrate, 'Parent', net.graph.sp5);
          set(net.graph.sp5_plot,'YDataSource','app.netCC.log.RCrate');
          grid;
          ylabel(net.graph.sp5,'RC rate');
          set(gca,'Xlim',[1 net.currentStep]);
          
          net.graph.sp6 = subplot(4,2,6,'Parent',net.graph.f1);
          
          if net.par.conditioningType == 1
              net.graph.sp6_plot1 = plot(net.log.stimulusToTrack,'Parent',net.graph.sp6,'Marker','o','MarkerFaceColor','blue');
              hold on
              net.graph.sp6_plot2 = plot(net.log.smile,'Parent',net.graph.sp6,'Marker','*','MarkerFaceColor','red');
              set(net.graph.sp6_plot1,'YDataSource','app.netCC.log.stimulusToTrack');
              set(net.graph.sp6_plot2,'YDataSource','app.netCC.log.smile');
              set(gca,'Ylim',[0.5 net.par.nrStimuli + 0.5]);
              ylabel(net.graph.sp6,'tracking stimulus');
              set(net.graph.f1,'Name','Classical Conditioning');
          else
              set(net.graph.f1,'Name','Operant Conditioning');
          end
          
          net.graph.sp7 = subplot(4,2,7,'Parent',net.graph.f1);
          ylabel(net.graph.sp7,'Raw stimuli');
          
          net.graph.sp8 = subplot(4,2,8,'Parent',net.graph.f1);
          ylabel(net.graph.sp8,'Processed stimuli');
          
      end
      
      function refreshGraphs(net)
          
          if net.par.conditioningType == 1
              net.graph.sp1_plot = plot(net.log.synapsesS(1:(net.currentStep-1),:),'Parent',net.graph.sp1);
              set(net.graph.sp1,'Xlim',[1 net.currentStep]);
              set(net.graph.sp1,'YGrid','on');
          else
              net.graph.sp1_plot = plot(net.log.synapsesS1O(1:(net.currentStep-1),:),'Parent',net.graph.sp1);
              set(net.graph.sp1,'Xlim',[1 net.currentStep]);
              set(net.graph.sp1,'YGrid','on');
          end
          
          set(net.graph.sp2,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          set(net.graph.sp3,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          
          
          net.graph.sp4_plot = plot(net.log.activityOG, 'Parent', net.graph.sp4);
          set(net.graph.sp4,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          set(net.graph.sp4,'YGrid','on');
          ylabel(net.graph.sp4,'output');
          
          set(net.graph.sp5,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          
          if net.par.conditioningType == 2
              net.graph.sp6_plot= plot(net.log.actions,'Parent',net.graph.sp6,'Line','none','Marker','.');
              set(net.graph.sp6,'Ylim',[0.5 net.par.nrOutputs + 0.5]);
              set(net.graph.sp6','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
              ylabel(net.graph.sp6,'Action');
          else
              set(net.graph.sp6,'YGrid','on');
              
          end
          
          set(net.graph.sp6,'Xlim',[max(1, net.currentStep - 1000) net.currentStep]);
          
          plot(net.log.stimuliRaw(:,:),'Parent',net.graph.sp7,'Marker','d','Line','--');
          set(net.graph.sp7,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          set(net.graph.sp7,'Ylim',[0.5 net.par.nrStimuli + 0.5]);
          set(net.graph.sp7','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
          
          plot(net.log.stimuliProcessed(:,:),'Parent',net.graph.sp8,'Marker','.','Line','--');
          set(net.graph.sp8,'Xlim',[max(1, net.currentStep - 500) net.currentStep]);
          set(net.graph.sp8,'Ylim',[0.5 net.par.nrStimuli + 0.5]);
          set(net.graph.sp8','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
          
          
          refreshdata(net.graph.f1);
          drawnow;
          
          %            pause(0.01);
      end
      
      function openWindowOC(net)
          net.graph.f1 = open('OCDemo.fig');
          net.graph.f1_subplots = get(net.graph.f1,'Children');
      end
      
      function refreshGraphsDemo(net)
          
          if net.par.conditioningType == 1
              set(net.graph.sp1,'Xlim',[1 net.currentStep-1]);
              set(net.graph.sp1,'YGrid','on');
              set(net.graph.sp2,'Xlim',[max(1, net.currentStep - 1000) net.currentStep - 1]);
          else
              plot(net.log.synapsesS1O(1:(net.currentStep-1),:),'Parent',net.graph.f1_subplots(1));
              set(net.graph.f1_subplots(1),'Xlim',[1 net.currentStep-1],'YGrid','on','YLim',[0 1]);
              
              plot(net.log.synapsesS2O(1:(net.currentStep-1),:),'Parent',net.graph.f1_subplots(3));
              set(net.graph.f1_subplots(3),'Xlim',[1 net.currentStep-1],'YGrid','on','YLim',[0 1]);
              
              plot(net.log.actions,'Parent',net.graph.f1_subplots(4),'Line','none','Marker','.');
              
              set(net.graph.f1_subplots(4),'Xlim',[max(1, net.currentStep - 1000) net.currentStep-1], 'Ylim',[0.5 net.par.nrOutputs + 0.5]);
              set(net.graph.f1_subplots(4)','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
              
          end
          
          plot(net.log.stimuliProcessed(:,:),'Parent',net.graph.f1_subplots(2),'Marker','.','Line','--');
          set(net.graph.f1_subplots(2),'Xlim',[max(1, net.currentStep - 1000) net.currentStep-1]);
          set(net.graph.f1_subplots(2),'Ylim',[0.5 net.par.nrStimuli + 0.5]);
          set(net.graph.f1_subplots(2)','Color',[0 0 0],'YGrid','on','YColor',[0.5 0.5 0.5]);
          
          
          refreshdata(net.graph.f1);
          
          drawnow;
      end
      
      function [] = initialiseConnectivity(net)
          net.connectivity = sparse(net.par.nrNeurons,net.par.nrNeurons);
          
          if net.par.regularConnectivity == 1
              t_axonCount = zeros(net.par.nrNeurons,1);
              for i = 1:net.par.nrNeurons
                  % disp(['Setting dendrides for neuron ' num2str(i)]);
                  t_avgExcitatoryDendrides = net.par.nrExcitatory * net.par.probConnect;
                  t_dendrides = net.myrandi(1,1, round(t_avgExcitatoryDendrides * 1.00), round(t_avgExcitatoryDendrides * 1.00));
                  t_fromExc = zeros(t_dendrides,1);
                  
                  for j = 1:t_dendrides
                      t_fromExc(j) = net.myrandi(1,1,1,net.par.nrExcitatory);
                      while size(find(t_fromExc == t_fromExc(j)),1) > 1 || t_fromExc(j) == j...
                              || t_axonCount(t_fromExc(j)) > ( net.par.nrNeurons * net.par.probConnect)
                          t_fromExc(j) = net.myrandi(1,1,1,net.par.nrExcitatory);
                      end
                  end
                  t_axonCount(t_fromExc) = t_axonCount(t_fromExc) + 1;
                  
                  t_avgInhibitoryDendrides = net.par.nrInhibitory * net.par.probConnect;
                  t_dendrides = net.myrandi(1,1,round(t_avgInhibitoryDendrides * 0.9), round(t_avgInhibitoryDendrides * 1.1));
                  t_fromInh = zeros(t_dendrides,1);
                  
                  for j = 1:t_dendrides
                      t_fromInh(j) = net.myrandi(1,1,net.par.nrExcitatory+1,net.par.nrNeurons);
                      while size(find(t_fromInh == t_fromInh(j)),1) > 1 || t_fromInh(j) == j...
                              || t_axonCount(t_fromInh(j)) > ( net.par.nrNeurons * net.par.probConnect)
                          t_fromInh(j) = net.myrandi(1,1,net.par.nrExcitatory+1,net.par.nrNeurons);
                      end
                  end
                  t_axonCount(t_fromInh) = t_axonCount(t_fromInh) + 1;
                  net.connectivity(t_fromExc,i) = 1;
                  net.connectivity(t_fromInh,i) = 1;
              end
          else
              net.connectivity = sparse((round(rand(net.stream,net.par.nrNeurons)+net.par.probConnect-0.5))...
                  .* (ones(net.par.nrNeurons) - diag(ones(1,net.par.nrNeurons))));
          end
          % Important: output neurons do not send signals back to the network, i.e. do
          % not contribute to recurrent connections.
          %net.connectivity(net.outputNeuronsGroups(:),:) = 0;
          % Important: input neurons do not receive inputs from the network
          %net.connectivity(:,net.stimuliNeuronsGroups(:)) = 0;
          
          % inhibitory neurons do not inhibit other inhibitory neurons
          %net.connectivity(net.par.nrExcitatory + 1:net.par.nrNeurons,net.par.nrExcitatory + 1:net.par.nrNeurons) = 0;
          % inhibitory neurons do not inhibit output neuron groups
          if net.par.conditioningType == 2
          %    net.connectivity(net.par.nrExcitatory + 1:net.par.nrNeurons,net.outputNeuronsGroups(:)) = 0;
          end
      end
      
      function [] = initialiseNeuronGroups(net)
          
          if 0 % here group neurons have random indexes
              for i = 1:net.par.nrStimuli
                  net.stimuliNeuronsGroups(i,:) = net.myrandi(net.stream, 1, net.par.nrNeuronsInGroup, 1, net.par.nrNeurons);
                  %check for uniqueness
                  currentGroup = net.stimuliNeuronsGroups(i,:);
                  j = 1;
                  while j <= net.par.nrNeuronsInGroup
                      isInGroup = (size(find(currentGroup == currentGroup(j)),2) > 1);
                      if isInS == 1
                          %    disp(['Neuron in S at position ' num2str(i) ' is duplicated... fixing']);
                          currentGroup(i) = net.myrandi(net.stream, 1, 1, 1, net.par.nrNeurons);
                          i = j - 1;
                      end
                      j = j + 1;
                  end
                  net.stimuliNeuronsGroups(i,:) = currentGroup;
              end
          else % here group neurons have adjacent indexes
              endIndex = 0;
%              nrInhib = net.par.nrNeuronsInGroup / 5;
              for i = 1:net.par.nrStimuli
                  startIndex = net.par.nrNeuronsInGroup * (i - 1) + 1;
                  endIndex = net.par.nrNeuronsInGroup * (i - 1) + net.par.nrNeuronsInGroup;
                  % the following four lines replace some neurons in the group with inhib
                  % neurons
 %                 startInhibIndex = nrInhib * (i - 1) + 801;
  %                endInhibIndex = nrInhib * (i - 1) + 800 + nrInhib;
                  indexes = startIndex:endIndex;
   %               indexes(net.par.nrNeuronsInGroup - nrInhib + 1:net.par.nrNeuronsInGroup) = startInhibIndex:endInhibIndex;
                  % end replacement
                  net.stimuliNeuronsGroups(i,:) = indexes;
              end
              
              startIndex = endIndex + 1;
            %  startInhibIndex = endInhibIndex + 1;
              for i = 1:net.par.nrOutputs
                  endIndex = startIndex + net.par.nrNeuronsOutGroup - 1;
             %     endInhibIndex = startInhibIndex + nrInhib - 1;
                  indexes = startIndex:endIndex;
              %    indexes(net.par.nrNeuronsOutGroup - nrInhib + 1:net.par.nrNeuronsOutGroup) = startInhibIndex:endInhibIndex;
                  net.outputNeuronsGroups(i,:) = indexes;
                  startIndex = endIndex + 1;
               %   startInhibIndex = endInhibIndex + 1;
              end
          end
      end
      
      function [values] = mysymrand(net, row, col, maxValue)
          values = rand(net.stream, row, col) .* 2 * maxValue - maxValue;
      end
      
      function[value] = myrandi(net, rows, cols, min_i, max_i)
          value = min_i + round( rand(net.stream,rows,cols) * (max_i-min_i+1) - 0.5);
      end
      
    end
    
end



